from ....core.auth import AuthorizationAuth
from ....core.action import ProfileDestinationAction
from ....core.http import HTTPStream
from ....utils.hrflow import generate_workflow_response

from pydantic import Field
from typing import Dict, Any, Iterator
import base64
import requests


class PushProfile(ProfileDestinationAction, HTTPStream):
    payload: Dict[str, Any] = dict()
    auth: AuthorizationAuth
    subdomain: str = Field(
        ...,
        description="Subdomain flatchr just before `flatchr.io`. For example subdomain=`my_subdomain.my` in "
        "`http://my_subdomain.my.flatchr.io/ABC`",
    )
    vacancy: str = Field(
        ...,
        description="The pool in which candidates will be placed. Findable in the URL",
    )

    def build_request_headers(self):
        super().build_request_headers()
        self.headers["content-type"] = "application/json"
        self.headers["Accept"] = "*/*"

    @property
    def base_url(self):
        return "https://{}.flatchr.io/vacancy/candidate/json".format(self.subdomain)

    @property
    def http_method(self):
        return "POST"

    def push(self, data):
        self.payload.clear()
        profile = next(data)
        self.payload.update(profile)
        response = self.send_request()
        if response.status_code >= 400:
            raise RuntimeError(
                "Push profile to flatchr failed : `{}`".format(response.content)
            )

    def format(self, data: Dict[str, Any]) -> Dict[str, Any]:
        def get_candidate_attachments(hrflow_profile):
            attachments_list = hrflow_profile.get("attachments")
            # We try to find 'resume' (which is 'original' after traitments).
            # If we don't find it, we search 'original'.
            fall_backs = ["resume", "original"]

            for file_name in fall_backs:
                for attachment in attachments_list:
                    if attachment["file_name"] == file_name:
                        # Get the base64's resume from its AWS url.
                        url = attachment["public_url"]
                        response = requests.get(url)
                        b64 = base64.b64encode(response.content)
                        return b64.decode()
            return None

        info = data.get("info")
        email = info.get("email") if info else None
        if email == None:
            raise Exception(
                f"No email for hrflow_profile {data.get('reference')} but one is mandatory"
            )

        profile = {
            "vacancy": self.vacancy,
            "firstname": info.get("first_name")
            if info.get("first_name") is not None
            else "N/A",
            "lastname": info.get("last_name")
            if info.get("last_name") is not None
            else "N/A",
            "type": "document",
            "email": email,
            "resume": {
                "fileName": "resume.pdf",
                "contentType": "application/octet-stream",
                "data": get_candidate_attachments(data),
            },
        }
        return profile


class EnrichProfile(ProfileDestinationAction, HTTPStream):
    payload: Dict[str, Any] = dict()
    auth: AuthorizationAuth
    subdomain: str = Field(
        ...,
        description="Subdomain flatchr just before `flatchr.io`. For example subdomain=`my_subdomain.my` in "
        "`http://my_subdomain.my.flatchr.io/ABC`",
    )
    vacancy: str = Field(
        ...,
        description="The pool in which candidates will be placed. Findable in the URL",
    )
    compagny: str = Field(
        ...,
        description="The id of the compagny",
    )

    def build_request_headers(self):
        super().build_request_headers()
        self.headers["content-type"] = "application/json"

    @property
    def base_url(self):
        return "https://{0}.flatchr.io/company/{1}/search/candidate".format(
            self.subdomain, self.compagny
        )

    @property
    def http_method(self):
        return "POST"

    def push(self, data):
        self.payload.clear()
        profile = next(data)
        self.payload.update(profile)
        response = self.send_request()
        if response.status_code >= 400:
            raise RuntimeError(
                "Push profile to flatchr failed : `{}`".format(response.content)
            )

    def format(self, data: Dict[str, Any]) -> Dict[str, Any]:
        def get_education_list(hr_flow_profile):
            hrflow_education_list = hr_flow_profile["educations"]
            flatchr_education_list = []

            for education in hrflow_education_list:
                location = education.get("location")
                flatchr_education = {
                    "institution": {
                        "communication": {
                            "address": [
                                {
                                    "formattedAddress": location.get("text")
                                    if location
                                    else None
                                }
                            ]
                        },
                        "name": education.get("school"),
                    },
                    "educationLevelCodes": [{"name": education.get("title")}],
                    "educationDegrees": [
                        {
                            "name": education.get("title"),
                            "date": education.get("date_start"),
                            "specializations": [],
                        }
                    ],
                    "end": education.get("date_end"),
                    "descriptions": [education.get("description")],
                }
                flatchr_education_list.append(flatchr_education)
            return flatchr_education_list

        def get_experiences_list(hr_flow_profile):
            hrflow_experiences_list = hr_flow_profile["experiences"]
            flatchr_experiences_list = []

            for experience in hrflow_experiences_list:
                location = experience.get("location")
                flatchr_education = {
                    "title": experience.get("title"),
                    "positionHistories": [
                        {
                            "organization": {
                                "communication": {
                                    "address": [
                                        {
                                            "countryCode": None,
                                            "city": None,
                                            "postalCode": None,
                                            "geoLocation": {
                                                "latitude": location.get("lat"),
                                                "longitude": location.get("lng"),
                                            },
                                            "formattedAddress": location.get("text")
                                            if location
                                            else None,
                                        }
                                    ]
                                },
                                "name": experience.get("company"),
                            },
                            "jobCategories": [],
                            "jobLevels": [],
                            "start": experience.get("date_start"),
                            "current": None,
                        }
                    ],
                    "start": experience.get("date_start"),
                    "current": None,
                    "descriptions": [experience.get("description")],
                }
                flatchr_experiences_list.append(flatchr_education)
            return flatchr_experiences_list

        def get_name():
            name = {"formattedName": "XXX XXXX", "given": "XXXX", "family": "XXXX"}
            return name

        def get_phone(hr_flow_profile):
            info = hr_flow_profile.get("info")
            phone = [
                {"dialNumber": info.get("phone") if info else None, "useCode": None}
            ]
            return phone

        def get_email(hr_flow_profile):
            info = hr_flow_profile.get("info")

            email = [{"address": info.get("email") if info else None}]
            return email

        def get_position(hr_flow_profile):
            hrflow_experiences_list = hr_flow_profile["experiences"]
            position = []

            for experience in hrflow_experiences_list:
                if experience.get("title") != None and experience.get("title") != "":
                    position.append(experience.get("title"))

            return position

        def get_employment_positions(hr_flow_profile):
            hrflow_experiences_list = hr_flow_profile["experiences"]
            employment_positions = []

            for experience in hrflow_experiences_list:
                if (
                    experience.get("company") != None
                    and experience.get("company") != ""
                ):
                    employment_positions.append(experience.get("company"))

            return employment_positions

        value = {
            "address": {
                "administrative_area_level_1": data["info"]["location"].get("text"),
                "location_lat": data["info"]["location"].get("lat"),
                "location_lng": data["info"]["location"].get("lng"),
            },
            "education": get_education_list(data),
            "employment": get_experiences_list(data),
            "name": get_name(),
            "phone": get_phone(data),
            "email": get_email(data),
            "position": get_position(data),
            "employment_positions": get_employment_positions(data),
            "experience": int(data.get("experiences_duration")),
        }

        profile = {
            "app_name": "HRMatch",
            "reference": data["info"].get("email"),
            "name": "parsing",
            "type": "applicants",
            "value": value,
        }

        return profile


import pytest

from hrflow_connectors.core.auth import OAuth2PasswordCredentialsBody
from hrflow_connectors.connectors.boards.crosstalent import GetAllJobs


@pytest.fixture
def auth(credentials):
    access_token_url = "https://test.salesforce.com/services/oauth2/token"

    auth = OAuth2PasswordCredentialsBody(
        access_token_url=access_token_url,
        client_id=credentials["crosstalent"]["oauth2"]["client_id"],
        client_secret=credentials["crosstalent"]["oauth2"]["client_secret"],
        username=credentials["crosstalent"]["oauth2"]["username"],
        password=credentials["crosstalent"]["oauth2"]["password"],
    )
    return auth


def test_Auth(auth):
    access_token = auth.get_access_token()
    assert isinstance(access_token, str)
    assert access_token != ""


def test_GetAllJobs(logger, auth, hrflow_client):
    action = GetAllJobs(
        auth=auth,
        subdomain="vulcain-eng--recette.my",
        hrflow_client=hrflow_client("dev-demo"),
        board_key="8eba188e1af123a9818d00974ff37b943b7d54f4",
        hydrate_with_parsing=True,
    )
    action.execute()


from hrflow_connectors.core.action import Action, BoardAction
import pytest
import requests
import responses
from hrflow import Hrflow


@pytest.fixture
def generated_data_list():
    list_to_filter = []
    list_to_filter.append(dict(element1="value1", element2="value2"))
    list_to_filter.append(dict(element1="value1", element2="value1"))
    list_to_filter.append(dict(element1="value2", element2="value1"))
    list_to_filter.append(dict(element1="value2", element2="value2"))
    return list_to_filter


def test_apply_logics_with_empty_logics_list(generated_data_list):
    action = Action()
    filtered_list = action.apply_logics(generated_data_list)

    assert len(filtered_list) == 4
    assert dict(element1="value1", element2="value2") in filtered_list
    assert dict(element1="value1", element2="value1") in filtered_list
    assert dict(element1="value2", element2="value1") in filtered_list
    assert dict(element1="value2", element2="value2") in filtered_list


def test_apply_logics_single_filter(generated_data_list):
    def filter_element1_with_value1(element):
        return element.get("element1") == "value1"

    action = Action(
        logics=["filter_element1_with_value1"],
        global_scope=globals(),
        local_scope=locals(),
    )
    filtered_list = list(action.apply_logics(generated_data_list))

    assert len(filtered_list) == 2
    assert dict(element1="value1", element2="value2") in filtered_list
    assert dict(element1="value1", element2="value1") in filtered_list


def test_apply_logics_two_filter(generated_data_list):
    def filter_element1_with_value1(element):
        return element.get("element1") == "value1"

    def filter_element2_with_value1(element):
        return element.get("element2") == "value1"

    action = Action(
        logics=["filter_element1_with_value1", "filter_element2_with_value1"],
        global_scope=globals(),
        local_scope=locals(),
    )
    filtered_list = list(action.apply_logics(generated_data_list))

    assert len(filtered_list) == 1
    assert dict(element1="value1", element2="value1") in filtered_list


def test_apply_logics_single_filter_without_interaction(generated_data_list):
    def filter_nothing(element):
        return True

    action = Action(
        logics=["filter_nothing"], global_scope=globals(), local_scope=locals()
    )
    filtered_list = list(action.apply_logics(generated_data_list))

    assert len(filtered_list) == 4
    assert dict(element1="value1", element2="value2") in filtered_list
    assert dict(element1="value1", element2="value1") in filtered_list
    assert dict(element1="value2", element2="value1") in filtered_list
    assert dict(element1="value2", element2="value2") in filtered_list


def test_extern_format_function():
    def extern_format(data):
        return dict(c=data["a"], d=data["a"] + data["b"])

    action = Action(
        format_function_name="extern_format",
        global_scope=globals(),
        local_scope=locals(),
    )

    job_to_transform = dict(a="aaa", b="bbb", f="fff")
    transformed_job = action.format_switcher(job_to_transform)
    assert transformed_job == dict(c="aaa", d="aaabbb")


def test_default_format_without_extern_format_function():
    action = Action(
        format_function_name=None,
    )
    job_to_transform = dict(a="aaa", b="bbb", f="fff")
    transformed_job = action.format(job_to_transform)
    assert transformed_job == dict(a="aaa", b="bbb", f="fff")


def test_overwritten_format_with_extern_format_function():
    def extern_format(data):
        return dict(c=data["a"], d=data["a"] + data["b"])

    class TestAction(Action):
        def format(self, data):
            return dict(f=data["f"], g=data["a"] + data["b"])

    action = TestAction(
        format_function_name="extern_format",
        global_scope=globals(),
        local_scope=locals(),
    )

    job_to_transform = dict(a="aaa", b="bbb", f="fff")
    transformed_job = action.format_switcher(job_to_transform)
    assert transformed_job == dict(c="aaa", d="aaabbb")


@responses.activate
def test_Action_connect_and_execute(generated_data_list):
    # Build a connector from `generated_data_list` to `http://test.test/push`
    class TestConnectorAction(Action):
        def pull(self):
            return generated_data_list

        def format(self, data):
            element1 = data.get("element1")
            element2 = data.get("element2")
            adapted_data = dict(
                element3=element1, element4="{}+{}".format(element1, element2)
            )
            return adapted_data

        def push(self, data):
            requests.post("http://test.test/push", json=dict(data=list(data)))

    # Expected Output
    expected_output = []
    expected_output.append(dict(element3="value1", element4="value1+value2"))
    expected_output.append(dict(element3="value1", element4="value1+value1"))
    expected_output.append(dict(element3="value2", element4="value2+value1"))
    expected_output.append(dict(element3="value2", element4="value2+value2"))

    # Mock requests
    # create a matcher to check if the JSON Body sent by the Connector is in the right shape and has the right values
    match = [responses.matchers.json_params_matcher(dict(data=expected_output))]

    responses.add(
        responses.POST,
        "http://test.test/push",
        status=200,
        match=match,
    )

    # Exec action
    action = TestConnectorAction()
    action.execute()


@pytest.fixture
def generated_jobs():
    def generate_jobs_page(page: int, jobs=30):
        job_list = []
        for i in range(jobs):
            job_list.append(dict(reference="{}-{}".format(i, page)))
        return job_list

    return generate_jobs_page


def generate_hrflow_search_response(data, max_page=2):
    meta = dict(maxPage=max_page)
    response = dict(code=200, message="Success", meta=meta, data=dict(jobs=data))
    return response


@responses.activate
def test_BoardAction_get_all_references_from_board(generated_jobs):
    # Generate pages of jobs
    page_1 = generated_jobs(page=1, jobs=30)
    page_2 = generated_jobs(page=2, jobs=29)

    # Generate responses return by Hrflow
    responses.add(
        responses.GET,
        "https://api.hrflow.ai/v1/jobs/searching?board_keys=%5B%22abc%22%5D&limit=30&page=1&sort_by=created_at",
        status=200,
        json=generate_hrflow_search_response(page_1, max_page=2),
    )

    responses.add(
        responses.GET,
        "https://api.hrflow.ai/v1/jobs/searching?board_keys=%5B%22abc%22%5D&limit=30&page=2&sort_by=created_at",
        status=200,
        json=generate_hrflow_search_response(page_2, max_page=2),
    )

    # Catch requests sent to Hrflow

    hrflow_client = Hrflow(api_user="", api_secret="")

    action = BoardAction(
        hrflow_client=hrflow_client, board_key="abc", hydrate_with_parsing=False
    )
    all_reference_iter = action.get_all_references_from_board()
    all_reference_list = list(all_reference_iter)

    assert len(all_reference_list) == 59

    all_reference_list.sort()
    page_1_expected = ["{}-{}".format(i, 1) for i in range(30)]
    page_2_expected = ["{}-{}".format(i, 2) for i in range(29)]

    reference_expected = page_1_expected + page_2_expected
    reference_expected.sort()

    assert all_reference_list == reference_expected


@pytest.fixture
def generated_parsing_text_response():
    text = "I love Python\ni speak english"
    ents = []
    ents.append(dict(start=7, end=13, label="HardSkill"))
    ents.append(dict(start=22, end=29, label="Language"))
    data = dict(ents=ents, text=text)
    response = dict(code=200, message="Success", data=data)
    return response


@responses.activate
def test_BoardAction_hydrate_job_with_parsing(generated_parsing_text_response):
    # Catch request
    responses.add(
        responses.POST,
        "https://api.hrflow.ai/v1/document/parsing",
        status=200,
        json=generated_parsing_text_response,
    )

    # Build Action
    hrflow_client = Hrflow(api_user="", api_secret="")

    action = BoardAction(
        hrflow_client=hrflow_client, board_key="abc", hydrate_with_parsing=False
    )
    section = dict(name="s", title=None, description="i speak english")
    job = dict(reference="REF123", summary="I love Python", sections=[section])

    assert len(job.get("skills", [])) == 0
    assert len(job.get("language", [])) == 0

    hydrated_job = action.hydrate_job_with_parsing(job)

    assert len(hydrated_job["skills"]) == 1
    assert hydrated_job["skills"][0] == dict(name="Python", type="hard", value=None)

    assert len(hydrated_job["languages"]) == 1
    assert hydrated_job["languages"][0] == dict(name="english", value=None)


@pytest.fixture
def generate_indexing_get_response():
    def indexing_get_response_func(code, message, archived_at):
        job = dict(key="klm", archived_at=archived_at)
        response = dict(code=code, message=message, data=job)
        return response

    return indexing_get_response_func


@responses.activate
def test_BoardAction_check_reference_in_board_for_job_not_in_board(
    generate_indexing_get_response,
):
    # Generate job
    job = dict(reference="REF1")

    # generated response
    message = "Invalid parameters. Unable to find object: job"
    generated_response = generate_indexing_get_response(
        code=400, message=message, archived_at=None
    )

    # Catch request
    responses.add(
        responses.GET,
        "https://api.hrflow.ai/v1/job/indexing?board_key=abc&reference=REF1",
        status=400,
        json=generated_response,
    )

    # Build Action
    hrflow_client = Hrflow(api_user="", api_secret="")

    action = BoardAction(
        hrflow_client=hrflow_client, board_key="abc", hydrate_with_parsing=False
    )
    check_response = action.check_reference_in_board(job)
    assert check_response


@responses.activate
def test_BoardAction_check_reference_in_board_for_not_archived_job_in_board(
    generate_indexing_get_response,
):
    # Generate job
    job = dict(reference="REF1")

    # generated response
    message = "Job details"
    generated_response = generate_indexing_get_response(
        code=200, message=message, archived_at=None
    )

    # Catch request
    responses.add(
        responses.GET,
        "https://api.hrflow.ai/v1/job/indexing?board_key=abc&reference=REF1",
        status=200,
        json=generated_response,
    )

    # Build Action
    hrflow_client = Hrflow(api_user="", api_secret="")

    action = BoardAction(
        hrflow_client=hrflow_client, board_key="abc", hydrate_with_parsing=False
    )
    check_response = action.check_reference_in_board(job)
    assert not check_response


@responses.activate
def test_BoardAction_check_reference_in_board_fail(generate_indexing_get_response):
    # Generate job
    job = dict(reference="REF1")

    # generated response
    message = "Fail"
    generated_response = generate_indexing_get_response(
        code=400, message=message, archived_at=None
    )

    # Catch request
    responses.add(
        responses.GET,
        "https://api.hrflow.ai/v1/job/indexing?board_key=abc&reference=REF1",
        status=400,
        json=generated_response,
    )

    # Build Action
    hrflow_client = Hrflow(api_user="", api_secret="")

    action = BoardAction(
        hrflow_client=hrflow_client, board_key="abc", hydrate_with_parsing=False
    )
    try:
        action.check_reference_in_board(job)
        assert False
    except RuntimeError:
        assert True


@responses.activate
def test_BoardAction_check_reference_in_board_for_archived_job_in_board_without_parsing(
    generate_indexing_get_response,
):
    # Generate job
    job = dict(reference="REF1")

    # generated response
    message = "Job details"
    generated_response = generate_indexing_get_response(
        code=200, message=message, archived_at="2021-12-25T00:00:00"
    )

    # Catch request
    responses.add(
        responses.GET,
        "https://api.hrflow.ai/v1/job/indexing?board_key=abc&reference=REF1",
        status=200,
        json=generated_response,
    )

    ## create a matcher to check if the JSON Body sent by the Connector is in the right shape and has the right values
    expected_body = dict(board_key="abc", reference="REF1", is_archive=False)
    match = [responses.matchers.json_params_matcher(expected_body)]
    responses.add(
        responses.PATCH,
        "https://api.hrflow.ai/v1/job/indexing/archive",
        status=200,
        match=match,
    )

    ## create a matcher to check if the JSON Body sent by the Connector is in the right shape and has the right values
    expected_body = dict(board_key="abc", key="klm", reference="REF1")
    match = [responses.matchers.json_params_matcher(expected_body)]
    responses.add(
        responses.PUT,
        "https://api.hrflow.ai/v1/job/indexing",
        status=200,
        match=match,
    )

    # Build Action
    hrflow_client = Hrflow(api_user="", api_secret="")

    action = BoardAction(
        hrflow_client=hrflow_client, board_key="abc", hydrate_with_parsing=False
    )
    check_response = action.check_reference_in_board(job)
    assert not check_response


@responses.activate
def test_BoardAction_check_reference_in_board_for_archived_job_in_board_with_parsing(
    generate_indexing_get_response, generated_parsing_text_response
):
    # Generate job
    section = dict(name="s", title=None, description="i speak english")
    job = dict(reference="REF1", summary="I love Python", sections=[section])

    assert len(job.get("skills", [])) == 0
    assert len(job.get("language", [])) == 0

    # generated response
    message = "Job details"
    generated_response = generate_indexing_get_response(
        code=200, message=message, archived_at="2021-12-25T00:00:00"
    )

    # Catch request
    responses.add(
        responses.GET,
        "https://api.hrflow.ai/v1/job/indexing?board_key=abc&reference=REF1",
        status=200,
        json=generated_response,
    )

    ## create a matcher to check if the JSON Body sent by the Connector is in the right shape and has the right values
    expected_body = dict(board_key="abc", reference="REF1", is_archive=False)
    match = [responses.matchers.json_params_matcher(expected_body)]
    responses.add(
        responses.PATCH,
        "https://api.hrflow.ai/v1/job/indexing/archive",
        status=200,
        match=match,
    )

    # Catch parsing request
    responses.add(
        responses.POST,
        "https://api.hrflow.ai/v1/document/parsing",
        status=200,
        json=generated_parsing_text_response,
    )

    ## create a matcher to check if the JSON Body sent by the Connector is in the right shape and has the right values
    hydrated_job = dict(job)
    hydrated_job["skills"] = [dict(name="Python", type="hard", value=None)]
    hydrated_job["languages"] = [dict(name="english", value=None)]
    hydrated_job["certifications"] = []
    hydrated_job["courses"] = []
    hydrated_job["tasks"] = []

    expected_body = dict(board_key="abc", key="klm", **hydrated_job)
    match = [responses.matchers.json_params_matcher(expected_body)]
    responses.add(
        responses.PUT,
        "https://api.hrflow.ai/v1/job/indexing",
        status=200,
        match=match,
    )

    # Build Action
    hrflow_client = Hrflow(api_user="", api_secret="")

    action = BoardAction(
        hrflow_client=hrflow_client, board_key="abc", hydrate_with_parsing=True
    )
    check_response = action.check_reference_in_board(job)

    assert not check_response


def test_BoardAction_get_all_references_from_stream():
    jobs_in_stream = [dict(reference="REF1"), dict(reference="REF2")]
    references_in_stream = ["REF1", "REF2"]

    class TestBoardAction(BoardAction):
        def pull(self):
            return jobs_in_stream

    # Build Action
    hrflow_client = Hrflow(api_user="", api_secret="")

    action = TestBoardAction(
        hrflow_client=hrflow_client,
        board_key="abc",
        hydrate_with_parsing=False,
        archive_deleted_jobs_from_stream=True,
    )
    references_got = action.get_all_references_from_stream()
    assert list(references_got) == references_in_stream


@responses.activate
def test_BoardAction_check_deletion_references_from_stream():
    references_in_stream = ["REF1", "REF2"]
    references_in_board = ["REF1", "REF4"]

    class TestBoardAction(BoardAction):
        def get_all_references_from_board(self):
            return references_in_board

        def get_all_references_from_stream(self):
            return references_in_stream

    # Catch Archive request
    expected_body = dict(board_key="abc", reference="REF4", is_archive=True)
    match = [responses.matchers.json_params_matcher(expected_body)]
    responses.add(
        responses.PATCH,
        "https://api.hrflow.ai/v1/job/indexing/archive",
        status=200,
        match=match,
    )

    # Build Action
    hrflow_client = Hrflow(api_user="", api_secret="")

    action = TestBoardAction(
        hrflow_client=hrflow_client, board_key="abc", hydrate_with_parsing=False
    )
    action.check_deletion_references_from_stream()


import datetime
from hrflow_connectors.utils.datetime_converter import (
    from_str_to_datetime,
    DateFormatError,
)


## YYYY/MM/DD
def test_simple_slash_date():
    got = from_str_to_datetime("2000/10/01")
    expected = datetime.datetime(2000, 10, 1, 0, 0)
    assert got == expected


### Separation : Space
def test_simple_slash_date_space_with_hours():
    got = from_str_to_datetime("2000/10/01 12")
    expected = datetime.datetime(2000, 10, 1, 12, 0)
    assert got == expected


def test_simple_slash_date_space_with_hours_minutes():
    got = from_str_to_datetime("2000/10/01 12:02")
    expected = datetime.datetime(2000, 10, 1, 12, 2)
    assert got == expected


def test_simple_slash_date_space_with_hours_minutes_seconds():
    got = from_str_to_datetime("2000/10/01 12:02:14")
    expected = datetime.datetime(2000, 10, 1, 12, 2, 14)
    assert got == expected


### Separation : T
def test_simple_slash_date_T_with_hours():
    got = from_str_to_datetime("2000/10/01T12")
    expected = datetime.datetime(2000, 10, 1, 12, 0)
    assert got == expected


def test_simple_slash_date_T_with_hours_minutes():
    got = from_str_to_datetime("2000/10/01T12:02")
    expected = datetime.datetime(2000, 10, 1, 12, 2)
    assert got == expected


def test_simple_slash_date_T_with_hours_minutes_seconds():
    got = from_str_to_datetime("2000/10/01T12:02:14")
    expected = datetime.datetime(2000, 10, 1, 12, 2, 14)
    assert got == expected


## YYYY-MM-DD
def test_simple_hyphen_date():
    got = from_str_to_datetime("2000-10-01")
    expected = datetime.datetime(2000, 10, 1, 0, 0)
    assert got == expected


### Separation : Space
def test_simple_hyphen_date_space_with_hours():
    got = from_str_to_datetime("2000-10-01 12")
    expected = datetime.datetime(2000, 10, 1, 12, 0)
    assert got == expected


def test_simple_hyphen_date_space_with_hours_minutes():
    got = from_str_to_datetime("2000-10-01 12:02")
    expected = datetime.datetime(2000, 10, 1, 12, 2)
    assert got == expected


def test_simple_hyphen_date_space_with_hours_minutes_seconds():
    got = from_str_to_datetime("2000-10-01 12:02:14")
    expected = datetime.datetime(2000, 10, 1, 12, 2, 14)
    assert got == expected


def test_simple_hyphen_date_space_with_hours_minutes_seconds_milliseconds_short():
    got = from_str_to_datetime("2000-10-01 12:02:14.123")
    expected = datetime.datetime(2000, 10, 1, 12, 2, 14, 123000)
    assert got == expected


def test_simple_hyphen_date_space_with_hours_minutes_seconds_milliseconds_long():
    got = from_str_to_datetime("2000-10-01 12:02:14.851541")
    expected = datetime.datetime(2000, 10, 1, 12, 2, 14, 851541)
    assert got == expected


### Separation : T
def test_simple_hyphen_date_T_with_hours():
    got = from_str_to_datetime("2000-10-01T12")
    expected = datetime.datetime(2000, 10, 1, 12, 0)
    assert got == expected


def test_simple_hyphen_date_T_with_hours_minutes():
    got = from_str_to_datetime("2000-10-01T12:02")
    expected = datetime.datetime(2000, 10, 1, 12, 2)
    assert got == expected


def test_simple_hyphen_date_T_with_hours_minutes_seconds():
    got = from_str_to_datetime("2000-10-01T12:02:14")
    expected = datetime.datetime(2000, 10, 1, 12, 2, 14)
    assert got == expected


def test_simple_hyphen_date_T_with_hours_minutes_seconds_milliseconds_short():
    got = from_str_to_datetime("2000-10-01T12:02:14.123")
    expected = datetime.datetime(2000, 10, 1, 12, 2, 14, 123000)
    assert got == expected


def test_simple_hyphen_date_T_with_hours_minutes_seconds_milliseconds_long():
    got = from_str_to_datetime("2000-10-01T12:02:14.851541")
    expected = datetime.datetime(2000, 10, 1, 12, 2, 14, 851541)
    assert got == expected


## Time zone
def test_simple_hyphen_date_T_with_hours_minutes_seconds_Z():
    got = from_str_to_datetime("2000-10-01T12:02:14Z")
    tz = datetime.timezone.utc
    expected = datetime.datetime(2000, 10, 1, 12, 2, 14, tzinfo=tz)
    assert got == expected


def test_simple_hyphen_date_T_with_hours_Z():
    got = from_str_to_datetime("2000-10-01T12Z")
    tz = datetime.timezone.utc
    expected = datetime.datetime(2000, 10, 1, 12, 0, tzinfo=tz)
    assert got == expected


def test_simple_hyphen_date_T_with_hours_minutes_seconds_milliseconds_long_plus_6_colons():
    got = from_str_to_datetime("2000-10-01T12:02:14.851541+06:00")
    tz = datetime.timezone(datetime.timedelta(hours=6))
    expected = datetime.datetime(2000, 10, 1, 12, 2, 14, 851541, tzinfo=tz)
    assert got == expected


def test_simple_hyphen_date_T_with_hours_minutes_seconds_milliseconds_long_plus_8():
    got = from_str_to_datetime("2000-10-01T12:02:14.851541+0800")
    tz = datetime.timezone(datetime.timedelta(hours=8))
    expected = datetime.datetime(2000, 10, 1, 12, 2, 14, 851541, tzinfo=tz)
    assert got == expected


def test_simple_hyphen_date_T_with_hours_minutes_seconds_moins_3():
    got = from_str_to_datetime("2000-10-01T12:02:14-0300")
    tz = datetime.timezone(datetime.timedelta(hours=-3))
    expected = datetime.datetime(2000, 10, 1, 12, 2, 14, tzinfo=tz)
    assert got == expected


## Random examples
def test_simple_date_with_colons_in_timezone():
    got = from_str_to_datetime("2021-10-11T14:57:33+00:00")
    tz = datetime.timezone.utc
    expected = datetime.datetime(2021, 10, 11, 14, 57, 33, tzinfo=tz)
    assert got == expected


def test_simple_date_in_HrFlow_tag():
    got = from_str_to_datetime("2021-10-13T10:57:38+0200")
    tz = datetime.timezone(datetime.timedelta(hours=2))
    expected = datetime.datetime(2021, 10, 13, 10, 57, 38, tzinfo=tz)
    assert got == expected


def test_simple_date_in_HrFlow():
    got = from_str_to_datetime("2021-10-01T12:59:05+0000")
    tz = datetime.timezone.utc
    expected = datetime.datetime(2021, 10, 1, 12, 59, 5, tzinfo=tz)
    assert got == expected


def test_wrong_date_with_letter():
    try:
        from_str_to_datetime("2021-1H-01T12:59:05+0000")
    except DateFormatError:
        pass


def test_wrong_date_with_invalid_date():
    try:
        from_str_to_datetime("2021-10-32T12:59:05+0000")
    except DateFormatError:
        pass


def test_wrong_date_with_invalid_format():
    try:
        from_str_to_datetime("2021-10sssss")
    except DateFormatError:
        pass


"""
    Conftest for `hrflow_connectors`
    
    All the fixtures used throughout the tests are grouped here.
"""

import pytest
import os
import json
import logging
from typing import Dict, Any, Callable
from hrflow import Hrflow
from webdriver_manager.chrome import ChromeDriverManager

from hrflow_connectors.utils.logger import get_logger_with_basic_config


@pytest.fixture(scope="session")
def credentials(pytestconfig) -> Dict[str, Any]:
    """
    Get credentials from a file in the root of the project `credentials.json` (to be defined)

    Returns:
        Dict[str, Any]: Credentials
    """
    with open(os.path.join(pytestconfig.rootpath, "credentials.json"), "r") as f:
        credentials = json.loads(f.read())
    return credentials


@pytest.fixture(scope="session")
def hrflow_client(credentials) -> Callable:
    """
    Get a function to generate an instance of Hrflow Client

    Returns:
        Callable: Function to generate an instance of Hrflow Client

    Example :
    >>> def test_stuff(hrflow_client):
    >>>     client = hrflow_client(portal_name="MY_PORTAL_ON_HRFLOW")
    """

    def hrflow_client_func(portal_name="dev-demo"):
        x_api_key = credentials["hrflow"][portal_name]["x-api-key"]
        x_user_email = credentials["hrflow"][portal_name]["x-user-email"]
        client = Hrflow(api_secret=x_api_key, api_user=x_user_email)
        return client

    return hrflow_client_func


@pytest.fixture(scope="session")
def webdriver_path() -> str:
    """
    Get the Chrome Web Driver path in using `webdriver_manager`

    Returns:
        str: Chrome Web Driver path
    """
    return ChromeDriverManager().install()


@pytest.fixture(scope="session")
def logger() -> logging.Logger:
    """
    Get `hrflow_connectors` Logger with basic config

    Returns:
        logging.Logger: `hrflow_connectors` Logger with basic config
    """
    return get_logger_with_basic_config()



from hrflow_connectors.utils.hrflow import find_element_in_list

def test_find_element_in_list():
    element_list = [dict(n="a", v=2), dict(n="b", v=3)]
    assert find_element_in_list(element_list, n="a") == dict(n="a", v=2)
    assert find_element_in_list(element_list, n="b", v=3) == dict(n="b", v=3)
    assert find_element_in_list(element_list, v=3) == dict(n="b", v=3)
    assert find_element_in_list(element_list, v=42) == None
    assert find_element_in_list(element_list, n=3) == None
    assert find_element_in_list(element_list, n="b", v=42) == None
    assert find_element_in_list(element_list, n="b", v=2) == None



import responses

from hrflow_connectors.connectors.boards.xml import XMLBoardAction
from hrflow_connectors.utils.datetime_converter import from_str_to_datetime


def test_get_all_jobs_from_samsic_xml_stream(logger, hrflow_client):
    def samsic_format(data):
        job = dict()

        # name
        job["name"] = data.findtext("Title")

        # reference
        job["reference"] = data.findtext("Reference")

        # created_at
        created_at_str = data.findtext("PublicationDateISO")
        created_at_datetime = from_str_to_datetime(created_at_str)
        job["created_at"] = created_at_datetime.isoformat()

        # url
        job["url"] = data.findtext(
            "PostingInstruction/ApplicationMethod/Communication/Text"
        )

        # location
        ## TODO ajouter utils de Zipcode -> Coords
        lat = None
        lng = None
        location_label_list = [
            "Location/CountryCode",
            "Location/City",
            "Location/PostalCode",
            "Location/Address",
        ]
        location_text_list = [
            data.findtext(label, default="") for label in location_label_list
        ]
        location_text = " ".join(location_text_list)

        geojson = dict()
        geojson["city"] = data.findtext("Location/City")
        geojson["country"] = data.findtext("Location/CountryCode")

        postcode = data.findtext("Location/PostalCode")

        geojson["postcode"] = postcode

        job["location"] = dict(lat=lat, lng=lng, text=location_text, geojson=geojson)

        # summary
        job["summary"] = data.findtext("Description")

        # sections
        job["sections"] = []

        def create_section(field_path: str, title: str = None):
            """
            Create a section in job if `field_name` value is not `None`

            Args:
                field_path (str): Field path used to retrieve value in the tree in Samsic. For example : `Organization/Description`
                title (str, optional): Section title. Defaults to None.
            """
            field_name = field_path.replace("/", "-")
            section_name = "samsic_{}".format(field_name)
            section_title = title
            section_description = data.findtext(field_path)
            if section_description is not None:
                section = dict(
                    name=section_name,
                    title=section_title,
                    description=section_description,
                )
                job["sections"].append(section)

        ## Add sections
        create_section("Organization/Description", "Organization description")
        create_section("Profile", "Profile")

        # languages
        job["languages"] = []

        # tags
        job["tags"] = []

        def create_tag(field_path: str):
            """
            Create tag in job if `field_path` is not `None`

            Args:
                field_path (str): Field path used to retrieve value in the tree in Samsic. For example : `Organization/Description`
            """
            field_name = field_path.replace("/", "-")
            tag_name = "crosstalent_{}".format(field_name)
            tag_value = data.findtext(field_path)
            if tag_value is not None:
                tag = dict(name=tag_name, value=tag_value)
                job["tags"].append(tag)

        ## Add tags
        create_tag("Duration")
        create_tag("Organization/Name")
        create_tag("Job")
        create_tag("Contract")
        create_tag("EducationRequirement")
        create_tag("LineOfBusiness")
        create_tag("FieldOfActivite")
        create_tag("DrivingLicence")
        create_tag("RequiredExperience")
        create_tag("Remuneration")

        start_date_str = data.findtext("StartDate")
        start_date_datetime = from_str_to_datetime(start_date_str)
        start_date_tag_value = start_date_datetime.isoformat()
        start_date_tag = dict(name="samsic_StartDate", value=start_date_tag_value)
        job["tags"].append(start_date_tag)

        # metadatas
        job["metadatas"] = []

        return job

    xml_stream_url = "https://cv.samsic-emploi.fr/media/flux/jobs.xml"
    job_list_xpath = "DataArea"

    action = XMLBoardAction(
        xml_stream_url=xml_stream_url,
        job_list_xpath=job_list_xpath,
        hrflow_client=hrflow_client("dev-demo"),
        board_key="ebf489eff6bef0e95ca03eb0d6ed8f8e030a634f",
        hydrate_with_parsing=False,
        archive_deleted_jobs_from_stream=True,
        format_function_name="samsic_format",
        global_scope=globals(),
        local_scope=locals(),
    )
    action.execute()


@responses.activate
def test_XMLBoardAction_pull_generic_xml_stream(logger, hrflow_client):
    xml_stream_url = "https://test.test/job/xml_stream"

    xml_stream_str = """<?xml version="1.0" encoding="UTF-8"?>
<all>
    <meta>
        <update>2021-11-09T15:00:00.999651</update>
        <version>42</version>
    </meta>
    <board>
        <jobs>
            <job data-id="1">
                <name>Data scientist</name>
                <reference>ds42</reference>
            </job>
            <job data-id="2">
                <name>Software Engineer</name>
                <reference>se18</reference>
            </job>
            <job data-id="3">
                <name>Painter</name>
                <reference>pt56</reference>
            </job>
        </jobs>
    </board>
</all>"""

    responses.add(responses.GET, xml_stream_url, status=200, body=xml_stream_str)

    job_list_xpath = "board/jobs"
    action = XMLBoardAction(
        xml_stream_url=xml_stream_url,
        job_list_xpath=job_list_xpath,
        hrflow_client=hrflow_client("dev-demo"),
        board_key="abc",
        hydrate_with_parsing=False,
        archive_deleted_jobs_from_stream=False,
    )
    xml_job_node_list = action.pull()

    assert len(xml_job_node_list) == 3
    job_node_1, job_node_2, job_node_3 = xml_job_node_list

    name_1, reference_1 = list(job_node_1)
    assert job_node_1.attrib["data-id"] == "1"
    assert name_1.text == "Data scientist"
    assert reference_1.text == "ds42"

    name_2, reference_2 = list(job_node_2)
    assert job_node_2.attrib["data-id"] == "2"
    assert name_2.text == "Software Engineer"
    assert reference_2.text == "se18"

    name_3, reference_3 = list(job_node_3)
    assert job_node_3.attrib["data-id"] == "3"
    assert name_3.text == "Painter"
    assert reference_3.text == "pt56"



from hrflow import Hrflow
import requests
import json


def workflow(settings):
    client = Hrflow(api_secret=settings["API_KEY"], api_user=settings["USER_EMAIL"])
    profiles = client.profile.searching.list(source_keys=["source_key"], limit = 20)['data']['profiles']
    for profile in profiles:
        jobs = client.job.scoring.list(board_keys=[settings["board_key"]], source_key=settings["source_key"], profile_key=profile['key'], agent_key=settings["agent_key"], use_agent=1, limit=3, order_by="desc", sort_by="scoring")['data']['jobs']
        sendGridData = {
            "personalizations": [
                {
                    "to": [
                        {
                            "email": profile['info']['email']
                        }
                    ],
                    "dynamic_template_data": {
                        'firstname': profile['info']['first_name'],
                        'jobtitle1': jobs[0]['name'],
                        'jobdescription1': jobs[0]['summary'],
                        'joburl1' : jobs[0]['url'],
                        'jobtitle2': jobs[1]['name'], 
                        'jobdescription2': jobs[1]['summary'],
                        'joburl2' : jobs[1]['url'],
                        'jobtitle3': jobs[2]['name'], 
                        'jobdescription3': jobs[2]['summary'],
                        'joburl3' : jobs[2]['url']
                    },
                    "subject": "Top3 hot jobs for you",
                }
            ],
            "from": {
                "email": "sender_email",
                "name": "sender_mail"
            },
            "template_id": "template_id",
            "content": [
                {
                    "type": "text/html"
                }
            ],
        }
        sendGridResp = requests.post("https://api.sendgrid.com/v3/mail/send", headers={'Authorization': 'Bearer '+settings["sendgrid_api_key"],'Content-Type': 'application/json'}, data=json.dumps(sendGridData)).content



from typing import Iterator, Dict, Any, Optional
from pydantic import Field
import itertools
import requests

from ....core.action import BoardAction
from ....core.http import HTTPStream
from ....core.auth import XSmartTokenAuth
from ....utils.logger import get_logger


logger = get_logger()


class SmartJobs(HTTPStream, BoardAction):
    auth: XSmartTokenAuth
    query: Optional[str] = Field(
        None,
        description="Full-text search query based on a job title; case insensitive; e.g. java developer",
    )
    updated_after: Optional[str] = Field(
        None, description="ISO8601-formatted time boundaries for the job update time"
    )
    posting_status: Optional[str] = Field(
        None,
        description="Posting status of a job. Available values : PUBLIC, INTERNAL, NOT_PUBLISHED, PRIVATE",
    )
    job_status: Optional[str] = Field(
        None,
        description="Status of a job. Available values : CREATED, SOURCING, FILLED, INTERVIEW, OFFER, CANCELLED, ON_HOLD",
    )
    limit: int = Field(
        10,
        description="Number of elements to return per page. max value is 100. Default value : 10",
    )

    @property
    def base_url(self):
        return "https://api.smartrecruiters.com/jobs"

    @property
    def http_method(self):
        return "GET"

    def pull(self) -> Iterator[Dict[str, Any]]:
        """
        Pull all jobs from SmartRecruiters

        Returns:
            Iterator[Dict[str, Any]]: an iterator of jobs
        """
        # If param value is `None`, `requests` in `HTTPStream` ignores the param
        self.params["q"] = self.query
        self.params["updatedAfter"] = self.updated_after
        self.params["postingStatus"] = self.posting_status
        self.params["status"] = self.job_status
        self.params["limit"] = self.limit

        # Define page generator
        def get_page():
            next_page_id = None
            job_list = None
            while job_list != []:
                self.params["pageId"] = next_page_id
                response = self.send_request()

                if response.status_code >= 400:
                    logger.error(f"Fail to get page of jobs : pageId=`{next_page_id}`")
                    error_message = "Unable to pull the data ! Reason : `{}`"
                    raise ConnectionError(error_message.format(response.content))

                logger.info(f"Get page of jobs : pageId=`{next_page_id}`")
                response_json = response.json()
                total_found = response_json["totalFound"]  # total jobs found
                next_page_id = response_json["nextPageId"]
                job_list = response_json["content"]
                job_number = len(job_list)
                logger.info(f"{job_number} job(s) got. Total found : {total_found}")
                yield job_list

        # Chain all jobs of each page in an Iterator
        chained_light_job_iter = itertools.chain.from_iterable(get_page())

        # `all_chained_job_iter` contains data-reduced jobs.
        # This is a feature of the `GET /jobs` search request.
        # We will retrieve all the information for each job by requesting it
        # from SmartRecruiter `GET /jobs/{jobId}`.

        def get_full_job(light_job_dict: Dict[str, Any]) -> Dict[str, Any]:
            """
            Get full job with all information available for the job  `light_job_dict`

            This function use `id` in `light_job_dict` to send request to SmartRecuiter and
            get all information of this job.

            Args:
                light_job_dict (Dict[str, Any]): light job

            Returns:
                Dict[str, Any]: full job
            """
            job_id = light_job_dict["id"]
            hearders = dict()
            self.auth.update(headers=hearders)
            get_job_url = f"https://api.smartrecruiters.com/jobs/{job_id}"
            response = requests.get(get_job_url, headers=hearders)

            if response.status_code >= 400:
                logger.error(f"Fail to get full job id=`{job_id}`")
                error_message = f"Unable to pull the job id=`{job_id}` ! Reason : `{response.content}`"
                raise ConnectionError(error_message)

            logger.info(f"Get full job id=`{job_id}`")
            return response.json()

        chained_full_job_iter = map(get_full_job, chained_light_job_iter)

        return chained_full_job_iter

    def format(self, data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Format a job from SmartRecruiter job format to Hrflow job format

        Args:
            data (Dict[str, Any]): Job offer

        Returns:
            Dict[str, Any]: Job in the HrFlow job object format
        """
        job = dict()

        # job Title
        job["name"] = data.get("title", "Undefined")

        # job Reference
        job["reference"] = data.get("refNumber")

        # job Url
        job["url"] = None

        # creation date and update date of the offer
        job["created_at"] = data.get("createdon")
        job["updated_at"] = data.get("updatedon")
        job["summary"] = None

        # location
        lat = data.get("location", {}).get("latitude")
        if lat is not None:
            lat = float(lat)

        lng = data.get("location", {}).get("longitude")
        if lng is not None:
            lng = float(lng)

        location_field_list = ["country", "region", "city", "address"]
        location_field_name_list = []
        for field_name in location_field_list:
            field = data.get("location", {}).get(field_name)
            if field is not None:
                location_field_name_list.append(field)

        text = " ".join(location_field_name_list)

        job["location"] = dict(lat=lat, lng=lng, text=text)

        # job sections: descriptions and qualifications
        job["sections"] = []
        job_ad_section_list = data.get("jobAd", {}).get("sections", {})

        def add_section(section_name: str):
            """
            Add section in Hrflow job

            Args:
                section_name (str): section name
            """
            name_prefix = "smartrecruiters_jobAd-sections"
            job_ad_section = job_ad_section_list.get(section_name)
            if job_ad_section is not None:
                name = f"{name_prefix}-{section_name}"
                title = job_ad_section.get("title")
                description = job_ad_section.get("text")
                section = dict(name=name, title=title, description=description)
                job["sections"].append(section)

        add_section("companyDescription")
        add_section("jobDescription")
        add_section("qualifications")
        add_section("additionalInformation")

        # job tags
        status = data.get("status")
        posting_status = data.get("postingStatus")
        job_id = data.get("id")
        experience_level_id = data.get("experienceLevel", {}).get("id")
        employmentType_id = data.get("typeOfEmployment", {}).get("id")
        industry_id = data.get("industry", {}).get("id")
        creator = data.get("creator", {})
        function_id = data.get("function", {}).get("id")
        department_id = data.get("department", {}).get("id")
        manual = data.get("location", {}).get("manual")
        remote = data.get("location", {}).get("remote")
        eeo_category_id = data.get("eeoCategory", {}).get("id")
        compensation = data.get("compensation", {})
        job["tags"] = [
            dict(name="smartrecruiters_status", value=status),
            dict(name="smartrecruiters_postingStatus", value=posting_status),
            dict(name="smartrecruiters_id", value=job_id),
            dict(name="smartrecruiters_experienceLevel-id", value=experience_level_id),
            dict(name="smartrecruiters_typeOfEmployment-id", value=employmentType_id),
            dict(
                name="smartrecruiters_compensation-min", value=compensation.get("min")
            ),
            dict(
                name="smartrecruiters_compensation-max", value=compensation.get("max")
            ),
            dict(
                name="smartrecruiters_compensation-currency",
                value=compensation.get("currency"),
            ),
            dict(name="smartrecruiters_industry-id", value=industry_id),
            dict(
                name="smartrecruiters_creator-firstName", value=creator.get("firstName")
            ),
            dict(
                name="smartrecruiters_creator-lastName", value=creator.get("lastName")
            ),
            dict(name="smartrecruiters_function-id", value=function_id),
            dict(name="smartrecruiters_department-id", value=department_id),
            dict(name="smartrecruiters_location-manual", value=manual),
            dict(name="smartrecruiters_location-remote", value=remote),
            dict(name="smartrecruiters_eeoCategory-id", value=eeo_category_id),
            dict(
                name="smartrecruiters_targetHiringDate",
                value=data.get("targetHiringDate"),
            ),
        ]

        return job



from hrflow import Hrflow
import mailchimp_transactional as MailchimpTransactional
from mailchimp_transactional.api_client import ApiClientError


def get_profile_event(_request: dict) -> dict:
    """
    Reconstruct webhook body
    @param _request: POST request
    @return: webhook event
    """
    return {
        "type": _request.get("type"),
        "origin": _request.get("type"),
        "message": _request.get("message"),
        "profile": _request.get("profile")
    }


def verify_webhook(signature: str, event: dict, secret: str) -> null:
    """
    Verify Webhook Signature
    @param signature: webhook integrity signature
    @param event: webhook data event
    @param secret: webhook secret key
    """
    import json
    import hmac
    import hashlib
    message = json.dumps(event, separators=(",", ":")).encode()
    hasher = hmac.new(secret.encode(), message, hashlib.sha256)
    dig = hasher.hexdigest()
    assert hmac.compare_digest(dig, signature)


def workflow(_request: dict, settings: dict) -> None:
    """
    WORKFLOW to send an email to a profile after catching a wehook parsing success event
    @rtype: null
    @param _request: dictionary that contains the body and the headers of the request
    @param settings: dictionary of settings params of the workflow
    """
    if not _request.get("profile"):
        return
    webhook_event = get_profile_event(_request)
    assert webhook_event["type"] == "profile.parsing.success"  # This workflow is only to create a new profile
    verify_webhook(_request["HTTP-HRFLOW-SIGNATURE"], webhook_event, settings["HRFLOW_WEBHOOK_SECRET"])
    profile_key = webhook_event["profile"]["key"]
    source_key = webhook_event["profile"]["source"]["key"]
    # Retrieve Profile from HrFlow.ai
    hrflow_client = Hrflow(api_secret=settings["HRFLOW_API_SECRET"], api_user=settings["HRFLOW_API_USER"])
    profile = hrflow_client.profile.indexing.get(source_key=source_key, key=profile_key).get("data")
    assert profile is not None
    # send Email with Mailchimp
    mailchimp = MailchimpTransactional.Client(settings["MANDRILL_API_TOKEN"])
    message = {
        "from_email": settings["HRFLOW_API_USER"],
        "subject": "Thank you for your application",
        "text": "Your application is well received. Our team will reach out to you if there are any opportunities "
                "matching your profile. Best, the Team",
        "to": [
            {
                "email": profile["info"]["email"],
                "type": "to"
            }
        ]
    }
    try:
        response = mailchimp.messages.send({"message": message})
        print("AcPI called successfully: {}".format(response))
    except ApiClientError as error:
        print("An excception occurred: {}".format(error.text))



from hrflow_connectors.connectors.boards.indeed import IndeedFeed


def test_IndeedFeed(logger, hrflow_client, webdriver_path):
    action = IndeedFeed(
        executable_path=webdriver_path,
        max_page=2,
        subdomain="fr",
        hrflow_client=hrflow_client("dev-demo"),
        job_search="Software Engineer",
        job_location="Paris",
        board_key="5865a71e45b94e29f7c1c97d71479ef2757df414",
        hydrate_with_parsing=True,
    )
    action.execute()



from hrflow_connectors.utils.hrflow import Profile, Source

def test_profile():
    profile = Profile(key="abc", source=Source(key="efg"))
    assert profile.key == "abc"
    assert profile.source.key == "efg"

def test_source():
    source = Source(key="efg")
    assert source.key == "efg"

def test_parse_profile():
    profile_dict = dict(key="abc", source=dict(key="efg"))
    profile_obj = Profile.parse_obj(profile_dict)
    assert profile_obj.key == "abc"
    assert profile_obj.source.key == "efg"


from hrflow_connectors.utils.timedelta_converter import (
    from_str_to_timedelta,
    TimeDeltaFormatError,
)


def assert_datetime(datetime_string):
    str_in_timedelta = from_str_to_timedelta(datetime_string)
    got_string = str(str_in_timedelta)
    expected_string = datetime_string
    assert got_string == expected_string


def test_simple_timedelta_with_8_hours():
    assert_datetime("8:05:32")


def test_simple_timedelta_with_23_hours():
    assert_datetime("23:05:32")


def test_simple_timedelta_with_days_hours():
    assert_datetime("66 days, 8:05:56")


def test_simple_neg_timedelta_with_days_hours():
    assert_datetime("-66 days, 8:05:56")


def test_simple_timedelta_with_days_hours_millisecond():
    assert_datetime("-1803 days, 15:54:03.999990")


def test_simple_neg_timedelta_with_days_hours_millisecond():
    assert_datetime("-1804 days, 15:54:03.999990")


def test_simple_neg_timedelta_with_one_day_hours_millisecond():
    assert_datetime("-1 day, 15:54:03.999990")


def test_wrong_timedelta_with_days_hours_millisecond_space_missing():
    try:
        assert_datetime("-1803days, 15:54:03.999990")
    except TimeDeltaFormatError:
        pass


def test_wrong_timedelta_with_days_hours_millisecond_with_letter():
    try:
        assert_datetime("-HH days, 15:54:03.999990")
    except TimeDeltaFormatError:
        pass


def test_wrong_timedelta_with_days_hours_millisecond_with_plus():
    try:
        assert_datetime("+01 days, 15:54:03.999990")
    except TimeDeltaFormatError:
        pass


def test_wrong_timedelta_with_days_hours_without_seconds():
    try:
        assert_datetime("1 days, 15:54")
    except TimeDeltaFormatError:
        pass


from typing import Dict, Any, List, Optional
from pydantic import BaseModel


def find_element_in_list(
    element_list: List[Dict[str, Any]], **fields
) -> Optional[Dict[str, Any]]:
    """
    Find element with some fields in list of elements

    >>> element_list = [dict(n="a", v=2), dict(n="b", v=3)]
    >>> find_element_in_list(element_list, n="a")
    ... dict(n="a", v=2)
    >>> find_element_in_list(element_list, n="b", v=3)
    ... dict(n="b", v=3)
    >>> find_element_in_list(element_list, v=42)
    ... None

    Args:
        element_list (List[Dict[str, Any]]): list of element with differents fields

    Returns:
        Dict[str, Any]: return element. Otherwise, return None.
    """
    for element in element_list:
        if element.items() >= fields.items():
            return element
    return None


def generate_workflow_response(status_code=201, **kwargs) -> Dict[str, Any]:
    """
    Generate CATCH workflow response

    Args:
        status_code (int, optional): status code like HTTP code. Defaults to 201.
        **kwargs: additional fields to add to the returned response

    Returns:
        Dict[str, Any]: Generated response
    """
    headers = {"Content-Type": "application/json"}
    response = dict(status_code=status_code, headers=headers)
    response.update(kwargs)
    return response


class Source(BaseModel):
    """
    Hrflow Source
    """

    key: str


class Profile(BaseModel):
    """
    Hrflow Profile
    """

    key: str
    source: Source


class EventParser(BaseModel):
    request: Dict[str, Any]

    def get_profile(self, source_to_listen: List[str] = None) -> Optional[Profile]:
        """
        Get profile

        If the parser does not find `Profile` object, then it will return `None`

        Args:
            source_to_listen (List[str], optional): list of source keys to listen. Defaults to None if you want to listen all available sources.

        Returns:
            Optional[Profile]: `Profile` object or `None` if not found
        """
        profile_dict = self.request.get("profile")
        if profile_dict is not None:
            profile_obj = Profile.parse_obj(profile_dict)
            if source_to_listen is None or profile_obj.source.key in source_to_listen:
                return profile_obj
        return None


from hrflow_connectors.utils.hrflow import generate_workflow_response


def test_generate_workflow_response_201_without_fields():
    response = generate_workflow_response(status_code=201)
    assert len(response) == 2
    assert response.get("status_code") == 201


def test_generate_workflow_response_400_without_fields():
    response = generate_workflow_response(status_code=400)
    assert len(response) == 2
    assert response.get("status_code") == 400


def test_generate_workflow_response_400_with_single_field():
    response = generate_workflow_response(status_code=400, field1="value1")
    assert len(response) == 3
    assert response.get("status_code") == 400
    assert response.get("field1") == "value1"


def test_generate_workflow_response_400_with_multi_fields():
    response = generate_workflow_response(
        status_code=400, field1="value1", field2="value2"
    )
    assert len(response) == 4
    assert response.get("status_code") == 400
    assert response.get("field1") == "value1"
    assert response.get("field2") == "value2"



from mimetypes import MimeTypes
import base64
from hrflow import Hrflow
import requests

def format_educations(educations):
    _educations = []

    for edu in educations:
        _edu = format_project(edu)
        _edu.update({
            "institution": edu.get("school") if edu.get("school") else "NaN",
            "degree": edu.get("title") if edu.get("title") else "NaN",
            "major": "NaN"
        })

        _educations.append(_edu)

    return _educations


def format_experiences(experiences):
    _experiences = []

    for exp in experiences:
        _exp = format_project(exp)
        _exp.update({
            "title": exp.get("title") if exp.get("title") else "NaN",
            "company": exp.get("company") if exp.get("company") else "NaN"
        })

        _experiences.append(_exp)

    return _experiences


def format_project(project):
    return {
        "current": False,
        "startDate": project.get("date_start").split("T")[0] if project.get("date_start") else "1980-01-01",
        "endDate": project.get("date_end").split("T")[0] if project.get("date_end") else "2100-01-01",
        "location": xstr(project.get("location").get("text")),
        "description": project.get("description")
    }


def format_resume(profile):
    _attachment = profile.get("attachments")
    mime = MimeTypes()
    for attachment in profile.get("attachments"):
        if attachment.get("type") == "resume":
            content = base64.b64encode(requests.get(attachment["public_url"]).content).decode("utf-8")
            mime_type = mime.guess_type(attachment["public_url"])[0]
            return {
                "fileName": attachment.get("file_name") + attachment.get("extension"),
                "mimeType": mime_type,
                "fileContent": content
            }


def hydrate(profile):
    info = profile.get("info")
    return {
        "firstName": info.get("first_name"),
        "lastName": info.get("last_name"),
        "email": info.get("email"),
        "phoneNumber": info.get("phone"),
        "location": {
            "country": xstr(info.get("location").get("fields").get("country")),
            "countryCode": "No",
            "region": xstr(info.get("location").get("fields").get("state")),
            "regionCode": "NaN",
            "city": xstr(info.get("location").get("fields").get("city")),
            "lat": info.get("location").get("lat") if info.get("location").get("lat") else 64.7805,
            "lng": info.get("location").get("lng") if info.get("location").get("lng") else -147.3694
        },
        "web": {
            "skype": "",
            "linkedIn": xstr(info.get("location").get("linkedin")),
            "facebook": xstr(info.get("location").get("facebook")),
            "twitter": xstr(info.get("location").get("twitter")),
            "website": ""
        },
        "tags": [],
        "education": format_educations(profile.get("educations")),
        "experience": format_experiences(profile.get("experiences")),
        "attachments": [],
        "consent": True
    }


def get_tag_value(tags, name):
    for tag in tags:
        if tag["name"] == name:
            return tag.get("value")
    return None


xstr = lambda s: s or " "


def workflow(body, settings):
    print("Welcome to catch: new application")

    profile_key = body["profile_key"]
    source_key = body["source_key"]

    job_key = body["job_key"]
    board_key = body["board_key"]

    token = settings["TOKEN"]

    headers = {
        "X-SmartToken": token
    }

    client = Hrflow(api_secret=settings["API_KEY"], api_user=settings["USER_EMAIL"])

    job = client.job.indexing.get(board_key=board_key, key=job_key).get("data")

    job_uuid = get_tag_value(job.get("tags"), "job_uuid")

    profile = client.profile.indexing.get(source_key=source_key, key=profile_key).get("data")

    smart_candidate = hydrate(profile)
    url = "https://api.smartrecruiters.com/jobs/" + job_uuid + "/candidates"
    smart_response = requests.post(url=url, json=smart_candidate, headers=headers)

    if not smart_response.ok:
        raise Exception("Invalid response from smartrecruiters API")

    profile_uuid = smart_response.json().get("id")

    url_attachment = "https://api.smartrecruiters.com/candidates/" + profile_uuid + "/jobs/" + job_uuid + "/attachments/"
    add_attachment_response = requests.post(url=url_attachment,
                                            headers=headers,
                                            data={
                                                "attachmentType": "RESUME"
                                            },
                                            files={"file": format_resume(profile).get("fileContent")})
    if not add_attachment_response.ok:
        raise Exception("Invalid response while adding Resume")




import datetime
import re
from typing import Union

class ParseError(ValueError):
    """
    Parse Error
    """
    def __init__(self, message):            
        super().__init__(message)

class DateFormatError(ParseError):
    """
    Date Format Error
    """
    def __init__(self, value):            
        super().__init__("Le format de la date ne respecte pas la norme ISO8601 : `{}`".format(value))


def any_to_int(any : Union[None, str]) -> int:
    """
    Convert int string to integer. if this value is `None`, return `0`
    Args:
        any (Union[None, str]): value to convert
    Returns:
        int: converted integer
    """
    if any is None:
        return 0
    return int(any)

def from_str_to_datetime(datetime_str: str) -> datetime.datetime:
    """
    Convert string to `datetime.datetime`.
    The date must respect ISO8601 format.
    Args:
        datetime_str (str): date formated with the norme ISO8601
    Raises:
        DateFormatError: "Le format de la date ne respecte pas la norme ISO8601 : ..."
    Returns:
        datetime.datetime: converted Datetime object
    """
    isoformat_regex = "^(?P<year>\d{4})(-|/)(?P<month>\d{2})(-|/)(?P<day>\d{2})(.(?P<hour>\d{2})(:(?P<minute>\d{2})(:(?P<second>\d{2})(\.(?P<millisecond>\d{1,6}))?)?)?)?(?P<tz>((?P<tz_symbol>-|\+)(?P<tz_hour>\d{2}):?(?P<tz_minute>\d{2})(:(?P<tz_second>\d{2})(\.(?P<tz_millisecond>\d{1,6}))?)?)|Z)?$"
    match = re.search(isoformat_regex, datetime_str)
    if not match:
        raise DateFormatError(datetime_str)

    # Extract value from string
    ## Datetime
    year, month, day = match.group("year", "month", "day")
    hour, minute, second, millisecond = match.group("hour", "minute", "second", "millisecond")

    ## TimeZone
    tz_field_in_str = match.group("tz")# example : +02:00
    tz_symbol = match.group("tz_symbol")
    tz_hour, tz_minute, tz_second, tz_millisecond = match.group("tz_hour", "tz_minute", "tz_second", "tz_millisecond")

    # Pad right with 0 the fields millisecond and tz_millisecond
    # "123" => "123000"
    if millisecond is not None:
        millisecond = millisecond.ljust(6, "0")
    if tz_millisecond is not None:
        tz_millisecond = tz_millisecond.ljust(6, "0")

    # Convert each field to int
    try:
        year, month, day = any_to_int(year), any_to_int(month), any_to_int(day)
        hour, minute, second, millisecond = any_to_int(hour), any_to_int(minute), any_to_int(second), any_to_int(millisecond)
        tz_hour, tz_minute, tz_second, tz_millisecond = any_to_int(tz_hour), any_to_int(tz_minute), any_to_int(tz_second), any_to_int(tz_millisecond)
    except ValueError:
        raise DateFormatError(datetime_str)

    # Convert to Object
    delta = datetime.timedelta(hours=tz_hour, minutes=tz_minute, seconds=tz_second, microseconds=tz_millisecond)
    if tz_symbol == "-":
        # when -HH:MM
        delta = -1 * delta
    
    timezone = datetime.timezone(delta)

    if tz_field_in_str is None:
        timezone = None
    
    try:
        return datetime.datetime(year=year, month=month, day=day, hour=hour, minute=minute, second=second, microsecond=millisecond, tzinfo=timezone)
    except ValueError:
        raise DateFormatError(datetime_str)


from hrflow_connectors.utils.clean_text import remove_html_tags


def test_remove_html_tag_with_br():
    html_content = "Hello <br>World !"
    output = remove_html_tags(html_content)
    expected = "Hello World !"
    assert output == expected


def test_remove_html_tag_without_tag_with_single_less_than_sign():
    html_content = "Hello < World !"
    output = remove_html_tags(html_content)
    expected = "Hello < World !"
    assert output == expected


def test_remove_html_tag_without_tag_with_multiple_less_than_signs():
    html_content = "Hello <<< World !"
    output = remove_html_tags(html_content)
    expected = "Hello <<< World !"
    assert output == expected


def test_remove_html_tag_with_div():
    html_content = 'Hello<div attr="One Two" attr2="Three"> World</div> !'
    output = remove_html_tags(html_content)
    expected = "Hello World !"
    assert output == expected



from ....core.auth import OAuth2PasswordCredentialsBody
from ....core.action import ProfileDestinationAction
from ....core.http import HTTPStream
from ....utils.hrflow import generate_workflow_response

from pydantic import Field
from typing import Dict, Any


class PushProfile(ProfileDestinationAction, HTTPStream):
    payload: Dict[str, Any] = dict()
    auth: OAuth2PasswordCredentialsBody
    subdomain: str = Field(
        ...,
        description="Subdomain Crosstalent just before `salesforce.com`. For example subdomain=`my_subdomain.my` in `http://my_subdomain.my.salesforce.com/ABC`",
    )

    def build_request_headers(self):
        super().build_request_headers()
        self.headers["content-type"] = "application/json"

    @property
    def base_url(self):
        return "https://{}.salesforce.com/services/apexrest/crta/HrFlowCreateProfile".format(
            self.subdomain
        )

    @property
    def http_method(self):
        return "POST"

    def format(self, data):
        firstname = data["info"].get("first_name")
        lastname = data["info"].get("last_name")
        email = data["info"].get("email")
        key = data["key"]

        # Case if lastname is not defined after CV parsing
        # This field is required for Crosstalent
        if lastname is None or lastname == "":
            data["info"]["last_name"] = "N/A"

        # Case if email is not defined after CV parsing
        # This field is required for Crosstalent
        if email is None or email == "":
            if firstname is not None and lastname is not None:
                email = f"{firstname}.{lastname}@vulcain.com"
            else:
                email = f"{key}@vulcain.com"
            data["info"]["email"] = email

        return data

    def push(self, data):
        self.payload.clear()
        profile = next(data)
        self.payload.update(profile)
        response = self.send_request()
        if response.status_code >= 400:
            raise RuntimeError(
                "Push profile to Crosstalent failed : `{}`".format(response.content)
            )



import requests
import responses
from typing import Dict

from hrflow_connectors.core.http import HTTPStream
from hrflow_connectors.core.auth import Auth, NoAuth


@responses.activate
def test_HTTPStream_get():
    request_url = "https://test.test/get"
    body = {"test": "hello"}

    # build Mock for request
    responses.add(responses.GET, request_url, status=200, json=body)

    # HTTPStream to test
    class TestHTTPStream(HTTPStream):
        auth: Auth = NoAuth()

        @property
        def http_method(self):
            return "GET"

        @property
        def base_url(self):
            return "https://test.test"

        def path(self):
            return "/get"

    action = TestHTTPStream()
    response = action.send_request()

    assert isinstance(response, requests.Response)
    assert response.status_code == 200
    assert "application/json" in response.headers.get("content-type")

    response_json = response.json()
    assert response_json.get("test") == "hello"


@responses.activate
def test_HTTPStream_post():
    request_url = "https://test.test/post"
    body = {"test": "hello"}

    # build Mock for request
    responses.add(responses.POST, request_url, status=200, json=body)

    # HTTPStream to test
    class TestHTTPStream(HTTPStream):
        auth: Auth = NoAuth()

        @property
        def http_method(self):
            return "POST"

        @property
        def base_url(self):
            return "https://test.test"

        def path(self):
            return "/post"

    action = TestHTTPStream()
    response = action.send_request()

    assert isinstance(response, requests.Response)
    assert response.status_code == 200
    assert "application/json" in response.headers.get("content-type")

    response_json = response.json()
    assert response_json.get("test") == "hello"


@responses.activate
def test_HTTPStream_get_with_header():
    request_url = "https://test.test/header"
    body = {"test": "hello"}

    # build Mock for request
    match = [responses.matchers.header_matcher(dict(d="world !"))]
    responses.add(responses.GET, request_url, status=200, json=body, match=match)

    # HTTPStream to test
    class TestHTTPStream(HTTPStream):
        auth: Auth = NoAuth()

        @property
        def http_method(self):
            return "GET"

        @property
        def base_url(self):
            return "https://test.test"

        def path(self):
            return "/header"

        def build_request_headers(self):
            self.headers["d"] = "world !"

    action = TestHTTPStream()
    response = action.send_request()

    assert isinstance(response, requests.Response)
    assert response.status_code == 200
    assert "application/json" in response.headers.get("content-type")

    response_json = response.json()
    assert response_json.get("test") == "hello"



import datetime
import re
from typing import Union


class ParseError(ValueError):
    """
    Parse Error
    """

    def __init__(self, message):
        super().__init__(message)


class TimeDeltaFormatError(ParseError):
    """
    TimeDelta Format Error
    """

    def __init__(self, value):
        super().__init__(
            "Le format de la durée ne respecte pas le format `[[-]DD days, ]HH:MM:SS[.mmmmmm]`. Example : `-1804 days, 15:54:03.999990` or `8:02:12` : `{}`".format(
                value
            )
        )


def any_to_int(any: Union[None, str]) -> int:
    """
    Convert int string to integer. if this value is `None`, return `0`
    Args:
        any (Union[None, str]): value to convert
    Returns:
        int: converted integer
    """
    if any is None:
        return 0
    return int(any)


def from_str_to_timedelta(timedelta_str: str) -> datetime.timedelta:
    """
    Convert string to `datetime.timedelta`.
    The date must respect the format `[[-]DD days, ]HH:MM:SS[.mmmmmm]`.
    Examples : `-1804 days, 15:54:03.999990` or `8:02:12`
    Args:
        datetime_str (str): formated timedelta
    Raises:
        TimeDeltaFormatError: Le format de la durée ne respecte pas le format `[[-]DD days, ]HH:MM:SS[.mmmmmm]`. Example : `-1804 days, 15:54:03.999990` or `8:02:12` : ..."
    Returns:
        datetime.timedelta: converted TimeDelta object
    """
    isoformat_regex = "^((?P<day>[-0-9]+) day(s)?, )?(?P<hour>\d+):(?P<minute>\d+):(?P<second>\d+)(\.(?P<millisecond>\d{1,6}))?$"
    match = re.search(isoformat_regex, timedelta_str)
    if not match:
        raise TimeDeltaFormatError(timedelta_str)

    # Extract value from string
    day, hour, minute, second, millisecond = match.group(
        "day", "hour", "minute", "second", "millisecond"
    )

    # Pad right with 0 the fields millisecond
    # "123" => "123000"
    if millisecond is not None:
        millisecond = millisecond.ljust(6, "0")

    # Convert each field to int
    try:
        day, hour, minute, second, millisecond = (
            any_to_int(day),
            any_to_int(hour),
            any_to_int(minute),
            any_to_int(second),
            any_to_int(millisecond),
        )
    except ValueError:
        raise TimeDeltaFormatError(timedelta_str)

    # Convert to Object
    try:
        return datetime.timedelta(
            days=day,
            hours=hour,
            minutes=minute,
            seconds=second,
            microseconds=millisecond,
        )
    except ValueError:
        raise TimeDeltaFormatError(timedelta_str)


from logging import log
import requests
from pydantic import BaseModel, Field
from typing import Optional, Dict, Any, Union, TypeVar


from .auth import Auth, NoAuth
from ..utils.logger import get_logger


logger = get_logger()


class HTTPStream(BaseModel):
    """
    HTTPStream
    Abstract Class

    To send a request with a JSON body, you must set the `content-type` header to `application/json`.
    It is important to write the header key `content-type` in lower case.
    """

    auth: Auth = NoAuth()

    params: Dict[str, str] = dict()
    headers: Dict[str, str] = dict()
    payload: Union[None, str, Dict[str, Any]] = None
    cookies: Dict[str, str] = dict()

    @property
    def base_url(self) -> Optional[str]:
        return None

    @property
    def http_method(self):
        return "GET"

    def path(self):
        return ""

    def build_request_params(self):
        pass

    def build_request_headers(self):
        pass

    def build_request_payload(self):
        pass

    def build_request_cookies(self):
        pass

    def get_content_type(self) -> Optional[str]:
        """
        Get Content-Type of the body request

        This function find the "Content-Type" in Header
        and it is NOT case sensitive.
        So, it can find "content-type" or "CONTENT-TYPE" as well.

        Returns:
            str: return content-type or None.
        """
        for key, value in self.headers.items():
            if key.lower() == "content-type":
                return value

    def send_request(self) -> requests.Response:
        """
        Send a HTTP request

        Returns:
            requests.Response: request response
        """
        if self.base_url is None:
            logger.error("Base URL `base_url` (property function) is not defined !")
            raise ConnectionError(
                "Base URL `base_url` (property function) is not defined !"
            )

        # Build the request
        url = self.base_url + self.path()
        logger.debug(f"URL use in HTTPStream : `{url}`")

        logger.debug("Preparing request...")
        logger.debug("Preparing request params")
        self.build_request_params()
        logger.debug("Preparing request headers")
        self.build_request_headers()
        logger.debug("Preparing request payload")
        self.build_request_payload()
        logger.debug("Preparing request cookies")
        self.build_request_cookies()
        logger.debug("The request has been prepared")

        # Add the auth property in the different sections
        logger.debug("Adding auth properties in the different request fields...")
        self.auth.update(
            params=self.params,
            headers=self.headers,
            cookies=self.cookies,
            payload=self.payload,
        )
        logger.debug("Auth properties have been added in the request fields")

        params = dict()
        params["method"] = self.http_method
        params["url"] = url
        params["params"] = self.params
        params["headers"] = self.headers
        params["cookies"] = self.cookies

        logger.debug(f"A `{self.http_method}` request will be sent to this URL `{url}`")

        # Check if request is a JSON application
        content_type = self.get_content_type()
        if (
            content_type is not None
            and "application/json" in content_type.lower()
            and isinstance(self.payload, dict)
        ):
            logger.debug("The payload has been detected as JSON")
            params["json"] = self.payload
        else:
            logger.debug("The payload has been detected as String or Dataform")
            params["data"] = self.payload

        logger.debug("Sending the request now")
        return requests.request(**params)



import requests
from zipfile import ZipFile
import io
import json
from collections import namedtuple
import requests as req
from datetime import datetime
from builtins import object
from hrflow import Hrflow


class Talentsoft:

    def __init__(self, client_url, client_id, client_secret):
        self.client_url = client_url
        self.client_id = client_id
        self.client_secret = client_secret
        self.headers = self.get_headers()

    def get_headers(self):
        body = "grant_type=client_credentials&scope=MatchingIndexation&client_id=" + str(self.client_id) + "&client_secret=" + str(
            self.client_secret)
        resp = req.post(self.client_url + "api/token",
                        headers={"Content-Type": "application/x-www-form-urlencoded"},
                        data=body)

        headers = {"Authorization": "bearer " + resp.json()["access_token"]}

        return headers

    def send_report(self, payload):
        resp = req.post(self.client_url + "api/exports/v1/reports",
                        headers=self.headers,
                        json=payload)
        return resp.ok

    def pull_vacancies(self, offset, limit):
        conditions = "filter=vacancyStatus::Published"
        resp = requests.get(self.client_url + "api/exports/v1/vacancies?offset=" + str(offset) + "&limit=" + str(
            limit) + "&" + conditions, headers=self.headers, stream=True)
        reader = StreamingReader(resp.content)
        for binary_array in reader.getFiles():
            zip = ZipFile(io.BytesIO(binary_array))
            data = json.loads(str(zip.read("offerdetail"), "utf-8"),
                              object_hook=lambda d: namedtuple("X", d.keys())(*d.values()))
            yield Vacancy(zip, data)



class Vacancy(object):
    def __init__(self, zip, data):
        self.zip = zip
        self.data = data

    @property
    def status(self):
        """
        status
        """
        return self.data.offerDetail.status

    @property
    def language_id(self):
        """
        languageId
        """
        return self.data.offerDetail.languageId

    @property
    def vacancy_category(self):
        """
        vacancyCategory
        """
        return self.data.offerDetail.vacancyCategory

    @property
    def vacancy_view(self):
        """
        vacancyView
        """
        return self.data.offerDetail.vacancyView

    @property
    def candidate_view(self):
        """
        candidateView
        """
        return self.data.offerDetail.candidateView

    @property
    def organisation(self):
        """
        organisation
        """
        return self.data.offerDetail.organisation

    @property
    def candidate_redirection_url(self):
        """
        candidateRedirectionUrl
        """
        return self.data.offerDetail.candidateRedirectionUrl

    @property
    def creation_date(self):
        """
        Datetime of sending time
        """
        try:
            creation_date = self.data.offerDetail.creationDate  # iso format
        except KeyError:
            creation_date = datetime.utcnow().isoformat()
        return creation_date

    @property
    def modifiction_date(self):
        """
        Datetime of sending time
        """
        try:
            modifiction_date = self.data.offerDetail.modificationDate  # iso format
        except KeyError:
            modifiction_date = datetime.utcnow().isoformat()
        return modifiction_date

    @property
    def requesting_party(self):
        """
        requestingParty
        """
        return self.data.offerDetail.requestingParty

    @property
    def location(self):
        """
        location
        """
        text = self.data.offerDetail.location.address
        return {
            "text": text,
            "lat": None,
            "lng": None
        }

    @property
    def criteria(self):
        """
        criteria
        """
        return self.data.offerDetail.criteria

    @property
    def job_description(self):
        """
        jobDescription
        """
        return self.data.offerDetail.jobDescription

    @property
    def languages(self):
        """
        languages
        """
        _languages = []
        for language in self.data.offerDetail.languages:
            _languages.append({"name": language.language.label, "value": language.languageLevel.label})
        return _languages

    @property
    def referral(self):
        """
        referral
        """
        return self.data.offerDetail.referral

    @property
    def published_on_intranet(self):
        """
        publishedOnIntranet
        """
        return self.data.offerDetail.publishedOnIntranet

    @property
    def published_on_internet(self):
        """
        publishedOnInternet
        """
        return self.data.offerDetail.publishedOnInternet

    @property
    def reference(self):
        """
        reference
        """
        return self.data.offerDetail.reference

    @property
    def general_information_custom_fields(self):
        """
        generalInformationCustomFields
        """
        return self.data.offerDetail.generalInformationCustomFields

    @property
    def custom_block1_custom_fields(self):
        """
        customBlock1CustomFields
        """
        return self.data.offerDetail.customBlock1CustomFields

    @property
    def custom_block2_custom_fields(self):
        """
        customBlock2CustomFields
        """
        return self.data.offerDetail.customBlock2CustomFields

    @property
    def custom_block3_custom_fields(self):
        """
        customBlock3CustomFields
        """
        return self.data.offerDetail.customBlock3CustomFields

    @property
    def custom_block4_custom_fields(self):
        """
        customBlock4CustomFields
        """
        return self.data.offerDetail.customBlock4CustomFields

    @property
    def contract_type(self):
        """
        contractType
        """
        return self.data.offerDetail.jobDescription.contractType

    @property
    def sections(self):
        """
        sections
        """

        customs = self.data.offerDetail.jobDescription.jobDescriptionCustomFields

        return [{"name": "description1",
                 "title": "description1",
                 "description": self.job_description.description1 or ""
                 },
                {"name": "description2",
                 "title": "description2",
                 "description": self.job_description.description2 or ""
                 },
                {"name": "Complément du descriptif",
                 "title": "Complément du descriptif",
                 "description": "\n".join([xstr(customs.longText1), xstr(customs.longText1), xstr(customs.longText1)])
                 }]

    @property
    def tags(self):
        """
        primaryProfile=X(id="_TS_metier_profil_RH", label="Ressources humaines")
        secondaryProfiles=[], country=X(id="_TS_CO_Country_France", label="France"),
        professionalCategory=X(id="_TS_CSP_ingenieur_et_cadre", label="Ingénieur & Cadre"))


        :return:
        """
        _tags = [
            {"name": "talentsoft-organisation-id", "value": self.organisation.id},
            {"name": "talentsoft-status-id", "value": self.status.id},
            {"name": "talentsoft-professionalCategory-id", "value": self.job_description.professionalCategory.id},
            {"name": "talentsoft-country-id", "value": self.job_description.country.id},
            {"name": "talentsoft-primaryProfile-id", "value": self.job_description.primaryProfile.id},
            {"name": "talentsoft-contractType-id", "value": self.contract_type.id},
            {"name": "talentsoft-publishedOnInternet", "value": self.published_on_internet},
            {"name": "talentsoft-publishedOnIntranet", "value": self.published_on_intranet}
        ]
        if self.criteria.experienceLevel:
            _tags.append({"name": "talentsoft-experienceLevel", "value": self.criteria.experienceLevel.id})

        if self.criteria.educationLevel:
            _tags.append({"name": "talentsoft-educationLevel", "value": self.criteria.educationLevel.id})

        return _tags

    @property
    def item(self):
        data = {
            "name": self.job_description.title,
            "reference": self.reference,
            "url": None,
            "summary": "",
            "location": self.location,
            "sections": self.sections,
            "ranges_float": [],
            "ranges_date": [],
            "skills": self.criteria.skills,
            "languages": self.languages,
            "labels": [],
            "tags": self.tags,
            "metadatas": [],
            "created_at": self.creation_date
        }
        return data

    def callback(self, status, message=None):
        # send reports to Talentsoft
        item = {
            "id": self.item.get("reference"),
            "date": self.item.get("created_at"),
            "lastUpdateDate": datetime.utcnow().isoformat(),
            "status": status,
        }
        if message:
            item["messages"] = [message]

        report = {
            "reportType": "Vacancies",
            "items": [item]
        }
        return report


class StreamingReader:
    def __init__(self, content):
        self._readingDataBlock = False
        self._content = content
        self._file = None
        self._resetChunk()
        self._resetDataBlock()

    def _resetDataBlock(self):
        self._currentZip = b""
        self._currentZipSize = 0

    def _resetChunk(self):
        self._currentChunk = None
        self._indexChunk = 0

    def getFiles(self):
        for self._currentChunk in [self._content]:
            while len(self._currentChunk) > self._indexChunk:
                if self._readingDataBlock:
                    self._readDataBlock()
                    if self._file is not None:
                        yield self._file
                        self._file = None
                else:
                    self._readSizeBlock()

    def _readDataBlock(self):
        sizeToRead = self._currentZipSize
        self._currentZip = self._currentChunk[self._indexChunk:self._indexChunk + sizeToRead]
        self._indexChunk += sizeToRead
        if True or len(self._currentZip) == self._currentZipSize:
            self._readingDataBlock = False
            self._file = self._currentZip
            self._resetDataBlock()

    def _readSizeBlock(self):
        self._currentZipSize = int.from_bytes(self._currentChunk[self._indexChunk:self._indexChunk + 4],
                                              byteorder="little")
        self._indexChunk += 4
        self._readingDataBlock = True

def format_skills(text: str, ents: list) -> list:
    """
    Get the list of skills according to the HrFlow.ai Job format
    @param text: text description of the job
    @param ents: list of entities in the text
    @return: list of skills
    """
    skills = [{ "name": text[ent["start"]:ent["end"]].lower(), "value": None,
                "type": "hard" if ent["label"] == "HardSkill" else "soft"}
              for ent in ents if ent["label"].endswith("Skill")]
    return list({v["name"]:v for v in skills}.values())

xstr = lambda s: s or ""

def workflow(settings: dict) -> None:
    """
    WORKFLOW to pull profiles
    @rtype: null
    @param settings: dictionary of settings params of the workflow
    """
    hrflow_client = Hrflow(api_secret=settings["API_KEY"], api_user=settings["USER_EMAIL"])

    talentsoft_client = Talentsoft(settings["BASE_URL"], settings["CLIENT_ID"], settings["CLIENT_SECRET"])

    hrflow_job = talentsoft_client.pull_vacancies(settings["OFFSET"], settings["LIMIT"])

    for job in hrflow_job:
       job_json = job.item
       verify_job = hrflow_client.job.indexing.get(board_key=settings["BOARD_KEY"], reference=job_json["reference"]).get("data")
       if not verify_job:
           try:
               job_json["agent_key"] = settings["AGENT_KEY"]
               # Parse skills
               SECTION_SEPARATOR = "\n\n"  # important to separate sections by double line jumps
               job_text = SECTION_SEPARATOR.join(section["description"] or "" for section in job_json["sections"])
               job_parsing = hrflow_client.document.parsing.post(text=job_text).get("data")
               job_json["skills"] = format_skills(job_text, job_parsing["ents"])
               # Save Job
               resp = hrflow_client.job.indexing.add_json(board_key=settings["BOARD_KEY"], job_json=job_json)

               if resp["code"] == 201:
                   print("Report sent for profile %s " % (job_json.get("reference")))
                   talentsoft_client.send_report(job.callback("success"))
           except requests.exceptions.RequestException:
               raise Exception("Saving job with reference %s failed" % (job_json["reference"]))



from hrflow import Hrflow
import requests


def workflow(settings: dict) -> None:
    """
    PULL WORKFLOW allows you to run the following code instructions on a regular basis

    @rtype: null
    @param settings: dictionary of settings params of the workflow
    """



from hrflow_connectors.connectors.boards.craigslist.actions import CraigslistFeed


def test_CraigslistJobs(logger, hrflow_client, webdriver_path):
    action = CraigslistFeed(
        executable_path=webdriver_path,
        subdomain="Paris",
        hrflow_client=hrflow_client("dev-demo"),
        board_key="fc197c369ed9a8d35041961b268076885b2ea6f2",
        hydrate_with_parsing=True,
    )
    action.execute()



from hrflow_connectors.utils.hrflow import EventParser, Profile

def test_event_with_profile():
    request = dict(profile=dict(key="abc", source=dict(key="efg")))
    event = EventParser(request=request)
    profile = event.get_profile()
    assert profile is not None
    assert isinstance(profile, Profile)
    assert profile.key == "abc"
    assert profile.source.key == "efg"

def test_event_without_profile():
    request = dict(other=dict(key="abc", source=dict(key="efg")))
    event = EventParser(request=request)
    profile = event.get_profile()
    assert profile is None

def test_event_with_profile_in_listened_source():
    request = dict(profile=dict(key="abc", source=dict(key="efg")))
    event = EventParser(request=request)
    source_to_listen = ["xyz", "efg"]
    profile = event.get_profile(source_to_listen=source_to_listen)
    assert profile is not None
    assert isinstance(profile, Profile)
    assert profile.key == "abc"
    assert profile.source.key == "efg"

def test_event_with_profile_not_in_listened_source():
    request = dict(profile=dict(key="abc", source=dict(key="efg")))
    event = EventParser(request=request)
    source_to_listen = ["xyz", "hij"]
    profile = event.get_profile(source_to_listen=source_to_listen)
    assert profile is None


import pytest

from hrflow_connectors.core.auth import OAuth2PasswordCredentialsBody
from hrflow_connectors.connectors.destinations.crosstalent import PushProfile
from hrflow_connectors.utils.hrflow import Profile, Source


@pytest.fixture
def auth(credentials):
    access_token_url = "https://test.salesforce.com/services/oauth2/token"

    auth = OAuth2PasswordCredentialsBody(
        access_token_url=access_token_url,
        client_id=credentials["crosstalent"]["oauth2"]["client_id"],
        client_secret=credentials["crosstalent"]["oauth2"]["client_secret"],
        username=credentials["crosstalent"]["oauth2"]["username"],
        password=credentials["crosstalent"]["oauth2"]["password"],
    )
    return auth


def test_PushProfile(logger, auth, hrflow_client):
    profile = Profile(
        key="ea5704b959c5e53aaef65c04ef5018ae1fee1a77",
        source=Source(key="15517d70b0870e4cf431eefd78f8b39cff5607e8"),
    )
    action = PushProfile(
        auth=auth,
        subdomain="vulcain-eng--recette.my",
        hrflow_client=hrflow_client("vulcain"),
        profile=profile,
    )
    response = action.execute()
    assert response.get("status_code") == 201



import os
from selenium import webdriver
from hrflow import Hrflow
import requests


class Crawler:
    """
    selenium Crawler Class
    """
    def __init__(self) -> object:
        chrome_options = webdriver.ChromeOptions()
        self._tmp_folder = "/tmp/chromium"
        if not os.path.exists(self._tmp_folder):
            os.makedirs(self._tmp_folder)
        if not os.path.exists(self._tmp_folder + "/user-data"):
            os.makedirs(self._tmp_folder + "/user-data")
        if not os.path.exists(self._tmp_folder + "/data-path"):
            os.makedirs(self._tmp_folder + "/data-path")
        if not os.path.exists(self._tmp_folder + "/cache-dir"):
            os.makedirs(self._tmp_folder + "/cache-dir")
        if not os.path.exists(self._tmp_folder + "/download-data"):
            os.makedirs(self._tmp_folder + "/download-data")
        self.download_location = self._tmp_folder + "/download-data"

        chrome_options.add_argument("--headless")
        chrome_options.add_argument("--no-sandbox")
        chrome_options.add_argument("--disable-gpu")
        chrome_options.add_argument("--window-size=1280x1696")
        chrome_options.add_argument("--user-data-dir={}".format(self._tmp_folder + "/user-data"))
        chrome_options.add_argument("--hide-scrollbars")
        chrome_options.add_argument("--enable-logging")
        chrome_options.add_argument("--log-level=0")
        chrome_options.add_argument("--v=99")
        chrome_options.add_argument("--single-process")
        chrome_options.add_argument("--data-path={}".format(self._tmp_folder + "/data-path"))
        chrome_options.add_argument("--ignore-certificate-errors")
        chrome_options.add_argument("--homedir={}".format(self._tmp_folder))
        chrome_options.add_argument("--disk-cache-dir={}".format(self._tmp_folder + "/cache-dir"))
        chrome_options.add_argument("--disable-dev-shm-usage")
        chrome_options.binary_location = "/opt/bin/headless-chromium"
        self._driver = webdriver.Chrome(chrome_options=chrome_options)

        print("Headless Chrome Initialized")
        params = {"behavior": "allow", "downloadPath": self._tmp_folder + "/download-data"}
        self._driver.execute_cdp_cmd("Page.setDownloadBehavior", params)

    def get_driver(self):
        return self._driver


def format_job(driver: object) -> dict:
    """
    Format the scrapped job according to the HrFlow.ai Job format
    @param driver: Crawler driver
    @return: job in the HrFlow.ai format of skills
    """
    location = driver.find_element_by_xpath("//*[@id='map']")
    tags = driver.find_element_by_xpath("//*[@class='attrgroup']").text.split("\n")
    
    return {
        "name": driver.find_element_by_xpath("//*[@id='titletextonly']").text,
        "agent_key": None,
        "reference": None,
        "url": driver.current_url,
        "created_at": driver.find_elements_by_xpath("//*[@class='postinginfo reveal']")[0].find_element_by_tag_name("time").get_attribute("datetime"),
        "updated_at": None,
        "summary": "",
        "location": {
            "text": None,
            "lat": location.get_attribute("data-latitude"),
            "lng": location.get_attribute("data-longitude")},
        "sections": [
            {"name": "description", "title": "Description", "description": driver.find_element_by_xpath("//*[@id='postingbody']").text}
        ],
        "skills": [],
        "languages": [],
        "tags": [
            {"name": "compensation", "value": tags[0].split(":")[1].strip()},
            {"name": "employment_type", "value": tags[1].split(":")[1].strip()}
        ],
        "ranges_date": [],
        "ranges_float": [],
        "metadatas": [],
    }


def format_skills(text: str, ents: list) -> list:
    """
    Get the list of skills according to the HrFlow.ai Job format
    @param text: text description of the job
    @param ents: list of entities in the text
    @return: list of skills
    """
    skills = [{ "name": text[ent["start"]:ent["end"]].lower(), "value": None,
                "type": "hard" if ent["label"] == "HardSkill" else "soft"}
              for ent in ents if ent["label"].endswith("Skill")]
    return list({v["name"]:v for v in skills}.values())


def workflow(settings: dict) -> None:
    """
    PULL WORKFLOW allows you to run the following code instructions on a regular basis

    WORKFLOW follows these steps:
        1- Init HrFlow.ai Client
        2- Open Jobboard URL
        3- Compute total_pages
        4- Iterate over all pages
             - Iterate over all jobs in the given page
                - For every job, check whether the job already indexed in HrFlow.ai"s board using Job API
                - If the job doesn"t exist :
                        - scrap and format the job
                        - enrich the job using HrFlow.ai Document API
                        - save the job using HrFlow.ai Job API

    @rtype: null
    @param settings: dictionary of settings params of the workflow
    """
    hrflow_client = Hrflow(api_secret=settings["API_KEY"], api_user=settings["USER_EMAIL"])
    c = Crawler()
    driver = c.get_driver()
    # Get MAX PAGES
    driver.get(settings["JOBBOARD_URL"])
    driver.maximize_window()
    total_jobs = int(driver.find_element_by_xpath("//*[@class='totalcount']").text)
    count_jobs = 120 # count jobs per Page
    total_pages = total_jobs // count_jobs + 1
    for page in range(0, total_pages):
        page_url = settings["JOBBOARD_URL"]+"s=%s"%((page+1)*count_jobs)
        driver.get(page_url)
        jobs = driver.find_elements_by_xpath("//*[@class='result-heading']")
        total_jobs = int(driver.find_element_by_xpath("//*[@class='totalcount']").text)
        for i in range(0, total_jobs):
            driver.get(jobs[i].find_element_by_tag_name("a").get_attribute("href"))
            reference = driver.find_element_by_xpath("//*[@class='postinginfo']").text.split(":")[0].strip()
            verify_job = hrflow_client.job.indexing.get(board_key=settings["BOARD_KEY"], reference=reference).get("data")
            if not verify_job:
                try:
                    job = format_job(driver)
                    job["reference"] = reference
                    job["agent_key"] = settings["AGENT_KEY"]
                    # Parse skills
                    SECTION_SEPARATOR = "\n\n"  # important to separate sections by double line jumps
                    job_text = SECTION_SEPARATOR.join(section["description"] or "" for section in job["sections"])
                    job_parsing = hrflow_client.document.parsing.post(text=job_text).get("data")
                    job["skills"] = format_skills(job_text, job_parsing["ents"])
                    # Save Job
                    hrflow_client.job.indexing.add_json(board_key=settings["BOARD_KEY"], job_json=job)
                except requests.exceptions.RequestException:
                    raise Exception("Saving job with reference %s failed"%(reference))



from typing import Dict, Any, Iterator, Optional
from pydantic import Field
from selenium import webdriver
from selenium.common.exceptions import WebDriverException
from re import search

from ....core.action import BoardAction
from ....utils.logger import get_logger

logger = get_logger()


class CraigslistFeed(BoardAction):
    archive_deleted_jobs_from_stream: bool = False
    subdomain: str = Field(
        ...,
        description="Subdomain just before `craigslist.org/d/emploi/search/jjj` for example subdomain = `paris` in `https://paris.craigslist.org/d/emploi/search/jjj`, it is also the localisation of the job offers",
    )
    executable_path: Optional[str] = Field(
        None,
        description="A separate executable that Selenium WebDriver used to control Chrome. Make sure you install the chromedriver with the same version as your local Chrome navigator",
    )

    binary_location: Optional[str] = Field(
        None,
        description="Location of the binary chromium, usually in HrFlow workflows it equals `/opt/bin/headless-chromium`",
    )

    jobs_per_page: int = Field(120, const=True)

    @property
    def base_url(self):

        return "https://{}.craigslist.org/d/emploi/search/jjj".format(self.subdomain)

    @property
    def Crawler(self):
        """
        Selenium Crawler function
        """
        logger.info("Configuring Chrome Webdriver...")
        chrome_options = webdriver.ChromeOptions()
        chrome_options.add_argument("--headless")
        chrome_options.add_argument("--no-sandbox")
        chrome_options.add_argument("--disable-dev-shm-usage")
        chrome_options.add_argument("--disable-gpu")
        chrome_options.add_argument("--window-size=1280x1696")
        chrome_options.add_argument("--hide-scrollbars")
        chrome_options.add_argument("--enable-logging")
        chrome_options.add_argument("--log-level=0")
        chrome_options.add_argument("--v=99")
        chrome_options.add_argument("--single-process")
        chrome_options.add_argument("--ignore-certificate-errors")

        if self.binary_location is not None:
            logger.info(f"Chrome binary location : {self.binary_location}")
            chrome_options.binary_location = self.binary_location

        if self.executable_path is None:
            driver = webdriver.Chrome(chrome_options)
        else:
            driver = webdriver.Chrome(
                executable_path=self.executable_path, chrome_options=chrome_options
            )

        return driver

    def pull(self) -> Iterator[str]:
        """
        Pull job links

        The role of this function is to scrap and retrieve each job offer link for each job posted

        Returns:
            Iterator[str]: list of scrapped job links
        """
        job_link_list = []

        driver = self.Crawler

        try:
            logger.info(f"Crawler get page : url=`{self.base_url}`")
            driver.get(self.base_url)
        except WebDriverException as e:
            logger.error(f"Fail to get page : url=`{self.base_url}`")
            logger.error(e)
            error_message = f"This site in not available, check if {self.subdomain}.craiglist.org is a valid subdomain"
            raise ConnectionError(error_message)

        driver.maximize_window()
        total_jobs = int(driver.find_element_by_xpath("//*[@class='totalcount']").text)
        logger.info(f"Total jobs to find : {total_jobs}")
        total_pages = total_jobs // self.jobs_per_page + 1
        logger.info(f"Total pages to crawl : {total_pages}")

        for page in range(0, total_pages):
            logger.info(f"Crawling page=`{page + 1}` ...")
            driver.get(self.base_url + "s=%s" % ((page + 1) * self.jobs_per_page))
            # jobs rows
            currant_page_job_list = driver.find_elements_by_xpath(
                "//*[@id='search-results']/li"
            )
            current_page_job_count = len(currant_page_job_list)
            logger.info(f"Number of jobs on this page : {current_page_job_count}")
            # retrieve job link from each job row
            job_link_list += [
                job.find_element_by_tag_name("a").get_attribute("href")
                for job in currant_page_job_list
            ]

        job_link_count = len(job_link_list)
        logger.info(f"Total job links found on all pages : {job_link_count}")
        return job_link_list

    def format(self, job_link: str) -> Dict[str, Any]:
        """
        Format job

        Generates a dictionary of a job attributes, for each job link the function scraps with selenium and parse useful attributes

        Args:
            job_link (str): job link parsed after the function pull is executed

        Returns:
            Dict[str, Any]: Job in the HrFlow job object format
        """
        job = dict()

        logger.debug("Extracting job from a web page")
        driver = self.Crawler

        logger.debug(f"Crawler go to `{job_link}`")
        driver.get(job_link)
        logger.debug(f"Scraping the web page...")

        # job name
        job["name"] = driver.find_element_by_xpath("//*[@id='titletextonly']").text

        # job reference
        # Searching for a match to get the reference to the job
        m = search("7(.+?).html", job_link)
        if m is not None:
            job["reference"] = m.group(1)

        # job url
        job["url"] = job_link

        # job create_at
        job["created_at"] = (
            driver.find_elements_by_xpath("//*[@class='postinginfo reveal']")[0]
            .find_element_by_tag_name("time")
            .get_attribute("datetime")
        )

        # job summary
        job["summary"] = None

        # job location
        job["location"] = dict(text=self.subdomain, lat=None, lng=None)

        # job sections
        job["sections"] = [
            dict(
                name="craigslist_description",
                title="Description",
                description=driver.find_element_by_xpath("//*[@id='postingbody']").text,
            )
        ]

        # tags
        tags = driver.find_elements_by_tag_name("b")
        job["tags"] = [
            dict(name="craigslist_compensation", value=tags[0].text),
            dict(name="craigslist_employment_type", value=tags[1].text),
        ]

        job["ranges_date"] = []
        job["ranges_float"] = []
        job["metadatas"] = []

        logger.debug(f"The web page has been scraped")
        return job


from typing import Dict, Any, Iterator, Optional
import re
from pydantic import Field
from selenium import webdriver
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from time import sleep
from selenium.common.exceptions import NoSuchElementException


from ....core.action import BoardAction
from ....utils.logger import get_logger


logger = get_logger()


class IndeedFeed(BoardAction):
    archive_deleted_jobs_from_stream: bool = False
    subdomain: str = Field(
        ...,
        description="Subdomain just before `indeed.com` for example subdomain =`fr` in `https:/fr.indeed.com`",
    )
    job_search: str = Field(
        ...,
        description="Name of the job position we want to search offers in `fr.indeed.com`",
    )
    job_location: str = Field(..., description="Location of the job offers")

    executable_path: Optional[str] = Field(
        None,
        description="A separate executable that Selenium WebDriver used to control Chrome. Make sure you install the chromedriver with the same version as your local Chrome navigator",
    )

    binary_location: Optional[str] = Field(
        None,
        description="Location of the binary chromium, usually in HrFlow workflows it equals `/opt/bin/headless-chromium`",
    )

    max_page: Optional[int] = Field(
        None, description="Maximum number of pages to search"
    )

    job_per_page: int = Field(15, const=True)

    @property
    def url_base(self) -> str:
        return "https:/{}.indeed.com".format(self.subdomain)

    @property
    def Crawler(self):
        """
        Selenium Crawler function
        """
        logger.info("Configuring Chrome Webdriver...")
        chrome_options = webdriver.ChromeOptions()
        chrome_options.add_argument("--headless")
        chrome_options.add_argument("--no-sandbox")
        chrome_options.add_argument("--disable-dev-shm-usage")
        chrome_options.add_argument("--disable-gpu")
        chrome_options.add_argument("--window-size=1280x1696")
        chrome_options.add_argument("--hide-scrollbars")
        chrome_options.add_argument("--enable-logging")
        chrome_options.add_argument("--log-level=0")
        chrome_options.add_argument("--v=99")
        chrome_options.add_argument("--single-process")
        chrome_options.add_argument("--ignore-certificate-errors")

        if self.binary_location is not None:
            logger.info(f"Chrome binary location : {self.binary_location}")
            chrome_options.binary_location = self.binary_location

        if self.executable_path is None:
            driver = webdriver.Chrome(chrome_options)
        else:
            logger.info(f"Webdriver executable path : {self.executable_path}")
            driver = webdriver.Chrome(
                executable_path=self.executable_path, chrome_options=chrome_options
            )

        return driver

    def path(self, job_offset: int) -> str:
        """
        Get the generated path

        This function generates the path and pagination of the job offers search

        Args:
            job_offset (int): Job offset. For example job_offset = 10 if we want to start from page 2

        Returns:
            str: return the generated path. For example it returns `/emplois?q=data%20scientist&l=paris&start=10` if we are looking for Data Scientist offers in Paris in page 2
        """
        return "/emplois?q={query}&l={location}&start={start}".format(
            query=self.job_search,
            location=self.job_location,
            start=job_offset,
        )

    def pull(self) -> Iterator[str]:
        """pull [the role of this function is to interact with indeed, click buttons and search offers based on job title and location.
        for each page we scrap all the job cards shown (usually 15 per page), and for each job card it retrieves its individual link]

        Returns:
            Iterator[str]: [list of scrapped jobLinks]
        """
        job_links_list = []
        driver = self.Crawler

        logger.info(f"Crawler go to `{self.url_base}`")
        driver.get(self.url_base)
        # Find the text box and enter the type of job the user wants to get offers data
        search = driver.find_element_by_id("text-input-what")
        logger.info(f"Crawler send a query `{self.job_search}`")
        search.send_keys(self.job_search)
        search.send_keys(Keys.RETURN)

        # Get the search count str for example 'Page 1 de 993 emplois'
        search_count_str = driver.find_element_by_id("searchCountPages").text
        # retrieve the number of total related job offers from string 'for example from 'Page 1 de 993 emplois' we get job_search_count = 993'
        search_count = (
            search_count_str.split()
        )  # split the string in a list of strings ['1', 'de', '993', 'emplois']
        start = search_count.index("de")
        end = search_count.index(
            "emplois"
        )  # find the words that surround the total number of jobs
        job_search_count = int(
            "".join([search_count[i] for i in range(start + 1, end)])
        )  # group different parts of the number together for example if it is 'Page 1 de 1 993 emplois' we get 1993

        logger.info(f"Number of jobs found for this query : {job_search_count}")
        # If there is only one job offer result -corresponding to 'emploi'- or none a value error is raised, it is a rare case and the custom should make sure to type in a real job name.

        total_page = int(
            job_search_count / self.job_per_page
        )  # for example for a total of 993 and a count of 15 we loop over 66 pages
        logger.info(f"Number of pages found for this query : {total_page}")

        if self.max_page is not None:
            logger.info(f"Max pages requests by the user : {self.max_page}")
            total_page = self.max_page

        for page in range(0, total_page):
            logger.info(f"Scraping the page `{page}` to get the list of jobs")
            page_url = self.url_base + self.path(job_offset=self.job_per_page * page)
            logger.info(f"Crawler go to `{page_url}`")
            driver.get(page_url)
            logger.info(f"Crawler sleep 3 secondes")
            sleep(3)
            try:  # get jobCards inside of a page
                logger.info("Getting mosaic of jobs")
                jobCards = WebDriverWait(driver, 10).until(
                    EC.presence_of_element_located((By.ID, "mosaic-provider-jobcards"))
                )
                logger.info("Mosaic of jobs has been found")

                # get joblink for each jobCard
                logger.info("Finding each job link")
                elements = jobCards.find_elements_by_class_name("tapItem")
                elements_count = len(elements)
                logger.info(f"Number of jobs found on this page : {elements_count}")
                for x in elements:
                    link = x.get_attribute("href")
                    job_links_list.append(link)
            finally:  # continue to next page
                pass

        job_link_count = len(job_links_list)
        logger.info(f"Number of total job links found : {job_link_count}")
        return job_links_list

    def format(self, job_link: str) -> Dict[str, Any]:
        """format [generates a dictionary of a job attributes, for each job link the function scraps with selenium and parse useful attributes.]

        Args:
            job_link (str): [job_link parsed after the function pull is executed. each job link is retrieved from the job_links_list scrapped in the pull function]

        Returns:
            Dict[str, Any]: [a job in the HrFlow job object format]
        """
        job = dict()

        logger.debug("Extracting job from a web page")
        driver = self.Crawler
        logger.debug(f"Crawler go to `{job_link}`")
        driver.get(job_link)
        logger.debug(f"Scraping the web page...")

        # name
        job["name"] = driver.find_element_by_class_name(
            "jobsearch-JobInfoHeader-title"
        ).text

        # reference
        m = re.search("jk=(.+?)&", job_link)
        if m:
            job["reference"] = m.group(1)
        else:
            job["reference"] = None

        # created_at, updated_at : isn't shown on ideed, TODO : convert "il y a n jours" into date time
        job["created_at"] = None
        job["updated_at"] = None

        # location
        # Sometimes there is a `COMPANY\nREVIEWS\nLOCATION'
        # Sometimes there is a `COMPANY\nLOCATION'
        subtitle_list = driver.find_element_by_class_name(
            "jobsearch-DesktopStickyContainer-subtitle"
        ).text.split("\n")

        if len(subtitle_list) >= 2:
            location = subtitle_list[-1]
        else:
            # In the case where there is only one item
            # It is assumed that the single element in the list is not a location
            location = "Undefined"

        job["location"] = dict(text=location, lat=None, lng=None)

        # url
        job["url"] = job_link

        # summary
        try:
            text = driver.find_element_by_id("jobDescriptionText").text
        except NoSuchElementException:
            text = None

        job["summary"] = text

        # description
        job["sections"] = [
            {"name": "description", "title": "description", "description": text}
        ]
        # compensation and jobType if they exist are in the same header so we get them both in one task of finding elements
        try:
            element = driver.find_element_by_class_name(
                "jobsearch-JobMetadataHeader-item"
            ).text
        except NoSuchElementException:
            element = "Null"

        # compensation
        # Need to be sure that we are parsing a salary and not a job Type because of indeed dynamic structure
        items = element.split()
        l = [
            "Temps plein",
            "Temps",
            "plein",
            "CDI",
            "plein,",
            "Stage",
            "CDD",
            "Apprentissage",
            "Alternance",
            "partiel",
        ]
        for item in l:
            if item in items:
                items.remove(item)
        salary = " ".join([items[i] for i in range(len(items))])

        # employment_type
        # avoid that Selenium parse useless and erronous information due to Indeed dynamic website architecture for example it can give jobType = "38 000 € par an - CDI" or "2 369 € - 4 645 € par mois - Temps plein, CDD"
        contract_type = ""
        if element not in [
            "CDI",
            "CDD",
            "Stage",
            "Alternance",
            "Temps plein",
            "Temps partiel",
            "Apprentissage",
            "Contrat Pro",
            "Stage, alternance",
        ]:
            if "CDI" in element.split():
                contract_type = "CDI"
            elif "plein" in element.split():
                contract_type = "CDI"
            elif "CDD" in element.split():
                contract_type = "CDD"
            else:
                contract_type = "null"
        jobType = contract_type

        job["tags"] = [
            dict(name="indeed_compensantion", value=salary),
            dict(name="indeed_employment_type", value=jobType),
        ]

        job["ranges_date"] = list()
        job["ranges_float"] = list()
        job["metadatas"] = list()

        driver.quit()
        logger.debug(f"The web page has been scraped")
        return job



import requests
from typing import Union, Dict, Optional
from pydantic import BaseModel, Field
from ..utils.logger import get_logger


logger = get_logger()


class Auth(BaseModel):
    """
    Auth

    This generic class will propose a common interface to all Auth
    """

    def update(
        self,
        params: Union[Dict[str, str], None] = None,
        headers: Union[Dict[str, str], None] = None,
        payload: Union[Dict[str, str], None] = None,
        cookies: Union[Dict[str, str], None] = None,
    ):
        """
        Updating fields with authentication information

        This class is responsible for placing the credentials in the right places

        Args:
            params (Union[Dict[str, str], None], optional): Params field. Defaults to None.
            headers (Union[Dict[str, str], None], optional): Headers field. Defaults to None.
            payload (Union[Dict[str, str], None], optional): Payload field. Defaults to None.
            cookies (Union[Dict[str, str], None], optional): Cookies field. Defaults to None.
        """
        pass


class NoAuth(Auth):
    """
    No Auth

    To be used if a class does not require authentication
    """

    pass


class OAuth2PasswordCredentialsBody(Auth):
    """
    OAuth2 by using a password and adding credentials in the body of the request used to get the "access token".
    """

    access_token_url: str
    client_id: str
    client_secret: str
    username: str
    password: str

    def get_access_token(self) -> str:
        """
        Get access token

        Returns:
            str: Access token
        """
        logger.debug("Getting the access token...")
        payload = dict()
        payload["grant_type"] = "password"
        payload["client_id"] = self.client_id
        payload["client_secret"] = self.client_secret
        payload["username"] = self.username
        payload["password"] = self.password

        logger.debug(
            f"Sending request to get access token (url=`{self.access_token_url}`)"
        )
        response = requests.post(self.access_token_url, data=payload)

        if response.status_code != 200:
            logger.error("OAuth2 failed for getting access token !")
            raise ConnectionError(
                "OAuth2 failed ! Reason : `{}`".format(response.content)
            )

        logger.debug("The access token has been got")
        return response.json()["access_token"]

    def update(
        self,
        params: Union[Dict[str, str], None] = None,
        headers: Union[Dict[str, str], None] = None,
        payload: Union[Dict[str, str], None] = None,
        cookies: Union[Dict[str, str], None] = None,
    ):
        if headers is not None:
            headers.update(
                {"Authorization": "OAuth {}".format(self.get_access_token())}
            )


class XAPIKeyAuth(Auth):
    """
    X-API-Key Auth

    Auth used to authenticate with an `X-API-Key` in the headers
    """

    name: str = Field(
        "X-API-KEY",
        description="Name of the API key that will be used as a header in the request. Default value : `X-API-KEY`",
    )
    value: str = Field(..., description="API key value")

    def update(
        self,
        params: Union[Dict[str, str], None] = None,
        headers: Union[Dict[str, str], None] = None,
        payload: Union[Dict[str, str], None] = None,
        cookies: Union[Dict[str, str], None] = None,
    ):
        if headers is not None:
            headers.update({self.name: self.value})


class AuthorizationAuth(XAPIKeyAuth):
    """
    Authorization Auth

    Auth used to authenticate with an API key named `Authorization` in the headers
    """

    name: str = Field("Authorization", const=True)


class XSmartTokenAuth(XAPIKeyAuth):
    """
    XSmart Token Auth

    Auth used to authenticate to SmartRecruiter with a token
    """

    name: str = Field("X-SmartToken", const=True)


import logging
from ..utils.clean_text import remove_html_tags
from ..utils.hrflow import find_element_in_list, Profile
from ..utils.hrflow import generate_workflow_response
from ..utils.logger import get_logger

from pydantic import BaseModel, Field
from typing import List, Dict, Any, Iterator, TypeVar, Optional, Union
import itertools
import xml.etree.ElementTree

Hrflow = TypeVar("Hrflow")
TalentDataType = Union[str, xml.etree.ElementTree.Element, Dict[str, Any]]

logger = get_logger()


class Action(BaseModel):
    logics: List[str] = Field(
        [], description="Function names to apply as filter before pushing the data"
    )
    global_scope: Optional[Dict[str, Any]] = Field(
        None, description="A dictionary containing the current scope's global variables"
    )
    local_scope: Optional[Dict[str, Any]] = Field(
        None, description="A dictionary containing the current scope's local variables"
    )

    format_function_name: Optional[str] = Field(
        None, description="Function name to format job before pushing"
    )

    workflow_catch: bool = Field(
        True,
        const=True,
        description="Indicates if the action is executable in a workflow catch",
    )
    workflow_pull: bool = Field(
        True,
        const=True,
        description="Indicates if the action is executable in a workflow pull",
    )

    def pull(self) -> Iterator[TalentDataType]:
        """
        Pull data
        """
        raise NotImplementedError("`pull` is not implemented")

    def apply_logics(self, data: Iterator[TalentDataType]) -> Iterator[TalentDataType]:
        """
        Apply filters defined in `logics` on the `data` stream

        Args:
            data (List[TalentDataType]): Data stream to filter

        Returns:
            List[TalentDataType]: Filtered data stream
        """
        filtered_list = data

        logic_count = len(self.logics)
        logger.info(f"Number of logic functions to apply : {logic_count}")
        if logic_count > 0:
            logger.info(f"Logic functions to apply : {self.logics}")

        for logic_function_name in self.logics:
            logger.info(
                f"Evaluating the logic function named `{logic_function_name}` ..."
            )
            logic_function = eval(
                logic_function_name, self.global_scope, self.local_scope
            )
            logger.info(
                f"The logic function named `{logic_function_name}` has been evaluated"
            )
            filtered_list = filter(logic_function, filtered_list)
        return filtered_list

    def format_switcher(self, data: TalentDataType) -> Dict[str, Any]:
        """
        Choose the right function to format the data and format the input data into a push-ready data schema.

        This function will :
        * Try to retrieve the external format function given in `format_function_name` and format the data with it
        * In the case where no external format function is given `format_function_name=None`,
          then the function will execute the internal format function `format` of the class.
          The `format` function can be overloaded to give it a defined behaviour.
        * If the `format` function of the class is not overloaded,
          the default behaviour of this function is to return the input data.

        WARNING: If you want to map the format function to rewrite a pipeline of an `Action` from a connector,
        you should use this function, not the `format` of the parent class.
        If you take `format` from the parent class, then `format_function_name` will be ignored
        and only `format` will be used to format the data.

        Args:
            data (TalentDataType): Data we want to adapt to the output format

        Returns:
            Dict[str, Any]: Data adapted to the input format of the push function, ready to be sent
        """
        if self.format_function_name is None:
            logger.debug("External format function is not defined")
            logger.debug("Using internal `format` function")
            return self.format(data)
        else:
            logger.debug(
                f"Evaluating the external format function `{self.format_function_name}` ..."
            )
            format_function = eval(
                self.format_function_name, self.global_scope, self.local_scope
            )
            logger.debug(
                f"The external format function `{self.format_function_name}` has been evaluated"
            )
            return format_function(data)

    def format(self, data: TalentDataType) -> Dict[str, Any]:
        """
        Format the input data into a push-ready data schema

        WARNING: If you want to map the format function to rewrite a pipeline of an `Action` from a connector,
        you should use the function `format_switcher`, not the `format` of the parent class.
        If you take `format` from the parent class, then `format_function_name` will be ignored
        and only `format` will be used to format the data.

        Args:
            data (TalentDataType): Data we want to adapt to the output format

        Returns:
            Dict[str, Any]: Data adapted to the input format of the push function, ready to be sent
        """
        return data

    def push(self, data: Iterator[Union[str, Dict[str, Any]]]):
        """
        Push data

        Args:
            data (List[Union[str, Dict[str, Any]]]): Data to push
        """
        raise NotImplementedError("`push` is not implemented")

    def execute(self) -> Optional[Dict[str, Any]]:
        """
        Execute action
        """
        logger.info("Start execution")

        logger.info("Pulling data...")
        input_data = self.pull()
        logger.info("Data has been pulled")

        logger.info("Applying logics...")
        filtered_data = self.apply_logics(input_data)
        logger.info("Logics have been applied")

        # connect each filtered_data to the format accepted by the pull function (destination, source, board)
        logger.info("Mapping format function...")
        output_data = map(self.format_switcher, filtered_data)
        logger.info("Format function has been mapped")

        logger.info("Pushing data...")
        self.push(output_data)
        logger.info("Data has been pushed")

        logger.info("All has been done for this connector !")


class BoardAction(Action):
    hrflow_client: Hrflow = Field(
        ...,
        description="Hrflow client instance used to communicate with the Hrflow.ai API",
    )
    board_key: str = Field(
        ..., description="Board key where the jobs to be added will be stored"
    )
    hydrate_with_parsing: bool = Field(False, description="Enrich the job with parsing")
    archive_deleted_jobs_from_stream: bool = Field(
        True,
        description="Archive Board jobs when they are no longer in the incoming job stream",
    )

    def get_all_job_pages_from_board(self) -> Iterator[Iterator[Dict[str, Any]]]:
        """
        Get all job pages from Board

        Yields:
            Iterator[Iterator[Dict[str, Any]]]: Iterator of job Iterator. For example: List of job list.
        """
        logger.info(f"Getting all job from Hrflow Board `{self.board_key}`")

        def get_jobs_page(page: int) -> Dict[str, Any]:
            logger.info(f"Getting page `{page}` from `{self.board_key}` board")
            response = self.hrflow_client.job.searching.list(
                board_keys=[self.board_key], limit=30, page=page
            )
            if response["code"] >= 400:
                raise RuntimeError(
                    "Hrflow searching failed : `{}`".format(response["message"])
                )
            return response

        job_page = get_jobs_page(page=1)
        max_page = job_page["meta"]["maxPage"]
        logger.info(f"Max page in board `{self.board_key}` : {max_page}")
        job_list = job_page["data"]["jobs"]
        job_count = len(job_list)
        logger.info(f"Number of jobs got in page `1` : {job_count}")
        yield job_list

        for page in range(2, max_page + 1):
            job_page = get_jobs_page(page=page)
            job_list = job_page["data"]["jobs"]
            job_count = len(job_list)
            logger.info(f"Number of jobs got in page `{page}` : {job_count}")
            if len(job_page["data"]["jobs"]) == 0:
                return
            yield job_list

    def get_all_references_from_board(self) -> Iterator[str]:
        """
        Get all job references from a Board

        Yields:
            Iterator[str]: Iterator with all job references
        """
        logger.info(f"Getting all references from Hrflow Board `{self.board_key}` ...")

        def get_reference_from_job_list(
            job_list: Iterator[Dict[str, Any]]
        ) -> Iterator[str]:
            """
            Get reference from job list

            Args:
                job_list (Iterator[Dict[str,Any]]): Job list

            Yields:
                Iterator[Dict[str, Any]]: Iteractor of references
            """
            get_reference_from_job = lambda job: job.get("reference")
            return map(get_reference_from_job, job_list)

        def none_filter(data: Optional[str]) -> bool:
            """
            Filter None type

            Args:
                data (Optional[str]): data to check

            Returns:
                bool: data is not None ?
            """
            return data is not None

        all_job_pages_iter = self.get_all_job_pages_from_board()
        logger.info("Mapping a function to extract references from list of job")
        all_reference_pages_iter = map(get_reference_from_job_list, all_job_pages_iter)
        logger.info("Chaining all pages of jobs")
        chain_reference_iter = itertools.chain.from_iterable(all_reference_pages_iter)
        logger.info("Filtering all `None` value in the chain")
        clean_iter = filter(none_filter, chain_reference_iter)
        logger.info("Iterator containing all references is ready !")
        return clean_iter

    def push(self, data: Iterator[Union[str, Dict[str, Any]]]):
        for job in data:
            reference = job.get("reference")
            logger.debug(
                f"Pushing a job ref=`{reference}` to Hrflow Board `{self.board_key}`"
            )
            response = self.hrflow_client.job.indexing.add_json(
                board_key=self.board_key, job_json=job
            )
            if response["code"] >= 400:
                message = response["message"]
                logger.error("Failed to push a job !")
                raise ConnectionError("Failed to push ! Reason : `{}`".format(message))

    def hydrate_job_with_parsing(self, job: Dict[str, Any]) -> Dict[str, Any]:
        """
        Hydrate job with parsing

        Enrich the different fields of the job by applying parsing to large texts like `summary` or `sections`

        Args:
            data (Dict[str, Any]): job to hydrate

        Returns:
            Dict[str, Any]: hydrated job
        """
        reference = job.get("reference")
        logger.debug(f"Hydrating the current job with parsing... (ref=`{reference}`)")

        # Concat `summary` text and each `section` together
        logger.debug("Concatenating job summary with job sections...")
        concatenated_str = job.get("summary")
        if concatenated_str is None:
            concatenated_str = ""

        section_list = job.get("sections")
        if section_list is not None:
            for section in section_list:
                section_description = section.get("description")
                if section_description is not None:
                    concatenated_str += "\n" + section_description
        logger.debug("Job summary with job sections have been concatenated")

        # Clean the `concatenated_str` by removing htlm tags
        logger.debug("Removing HTML tags in concatenated texts...")
        cleaned_str = remove_html_tags(concatenated_str)
        logger.debug("HTML tags in concatenated texts have been removed")

        # If text is empty, the parsing can return an error
        if cleaned_str == "":
            logger.debug("Cleaned text is empty")
            return job

        # Parse the `cleaned`
        logger.debug("Parsing the cleaned text...")
        response = self.hrflow_client.document.parsing.post(text=cleaned_str)
        if response["code"] >= 400:
            logger.error("Fail to parse the cleaned text !")
            raise RuntimeError("Parsing failed : `{}`".format(response["message"]))
        logger.debug("Text has been parsed")

        entity_list = response["data"]["ents"]
        parsed_text = response["data"]["text"]
        entity_count = len(entity_list)
        logger.debug(f"Number of entity found by parsing : {entity_count}")

        # Enrich job with parsing
        logger.debug("Enriching the current job with parsing")
        ## Initialize each field that can be enrich (case of None type)
        logger.debug("Initializing `None` fields to enrich")
        field_to_init = ["skills", "languages", "certifications", "courses", "tasks"]
        for field_name in field_to_init:
            field_value = job.get(field_name)
            if field_value is None:
                logger.debug(
                    f"Field `{field_name}` has been initialized with an empty list"
                )
                job[field_name] = []

        ## Map parsing labels with job fields
        logger.debug("Mapping parsing label with job fields")
        label_to_job_field = dict(
            Course="courses",
            Task="tasks",
            Certification="certifications",
            Language="languages",
        )
        for entity in entity_list:
            start = entity["start"]
            end = entity["end"]
            label = entity["label"]
            selection = parsed_text[start:end]

            if label in ["Course", "Task", "Certification", "Language"]:
                element_to_add = dict(name=selection, value=None)
                # If element to add is unique. To avoid doublon.
                if (
                    find_element_in_list(
                        element_list=job[label_to_job_field[label]], name=selection
                    )
                    is None
                ):
                    logger.debug(
                        f"Add the field `{label_to_job_field[label]}` with the name `{selection}`"
                    )
                    job[label_to_job_field[label]].append(element_to_add)
                else:
                    logger.debug(
                        f"The field `{label_to_job_field[label]}` with the name `{selection}` is already added"
                    )
            elif label in ["Skill", "HardSkill", "SoftSkill"]:
                label_to_skill_type = dict(
                    Skill=None, HardSkill="hard", SoftSkill="soft"
                )
                skill_type = label_to_skill_type[label]
                skill = dict(name=selection, type=skill_type, value=None)
                # If element to add is unique. To avoid doublon.
                if (
                    find_element_in_list(element_list=job["skills"], name=selection)
                    is None
                ):
                    logger.debug(
                        f"Add the skill `{skill_type}` with the name `{selection}`"
                    )
                    job["skills"].append(skill)
                else:
                    logger.debug(
                        f"The skill `{skill_type}` with the name `{selection}` is already added"
                    )

        logger.debug("The job has been enriched !")
        return job

    def check_reference_in_board(self, job: Dict[str, Any]) -> bool:
        """
        Check if a job reference is in the Board.
        If the job reference is not in the Board, return `True` to add the job.
        Otherwise return `False` and if job is archived, this function unarchives it.

        Args:
            job (Dict[str, Any]): job object

        Returns:
            bool: Job is in the Board
        """
        reference = job.get("reference")
        logger.debug(
            f"Checking if the reference `{reference}` is already in Hrflow Board `{self.board_key}`"
        )

        if reference is None:
            logger.debug("This job reference is `None`. Thus, it is not in the Board")
            return True

        logger.debug("Sending a GET Indexing request to Hrflow...")
        response = self.hrflow_client.job.indexing.get(
            board_key=self.board_key, reference=reference
        )
        response_code = response["code"]
        if response_code >= 400 and "Unable to find object: job" in response["message"]:
            # Job with this reference is not in the Board
            logger.debug("Hrflow does not find this job reference")
            return True
        elif response_code >= 400:
            logger.error("GET Indexing failed !")
            raise RuntimeError("GET Indexing failed : {}".format(response["message"]))
        elif response_code == 200:
            logger.debug("Hrflow got this job in the Board")
            job_in_board = response["data"]
            archived_at = job_in_board.get("archived_at")
            if archived_at is None:
                # Job is not archived
                logger.debug("Job is not archived")
                return False
            else:
                # Job is archived
                logger.debug(f"Job is archived since `{archived_at}`")
                logger.debug("Un-archiving the job in Hrflow")
                archive_response = self.hrflow_client.job.indexing.archive(
                    self.board_key, reference=reference, is_archive=0
                )
                if archive_response["code"] >= 400:
                    error_message = archive_response["message"]
                    logger.warning(
                        f"Fail to un-archive the job `{reference}` : {error_message}"
                    )
                else:
                    # Get job key
                    job_response = response["data"]
                    job_key = job_response["key"]

                    # Hydrate job with parsing if it is necessary
                    if self.hydrate_with_parsing:
                        job = self.hydrate_job_with_parsing(job)

                    # Edit job
                    logger.debug(f"Editing the job key=`{job_key}`")
                    edit_response = self.hrflow_client.job.indexing.edit(
                        board_key=self.board_key, key=job_key, job_json=job
                    )
                    if edit_response["code"] >= 400:
                        error_message = edit_response["message"]
                        logger.warning(
                            f"Fail to edit the job `{reference}` : {error_message}"
                        )
                return False
        return False

    def get_all_references_from_stream(self) -> Iterator[str]:
        """
        Get all job references from stream

        Yields:
            Iterator[str]: return all references
        """
        logger.info(f"Getting all references from the stream...")
        logger.info(f"Pulling all references from the stream")
        input_data = self.pull()
        logger.info(f"Applying logics to all references from the stream")
        filtered_data = self.apply_logics(input_data)
        logger.info(f"Formating all references from the stream")
        formated_data = map(self.format_switcher, filtered_data)
        logger.info(f"Keeping only reference from the stream")
        references_iter = map(lambda job: job.get("reference"), formated_data)
        references_without_none_iter = filter(
            lambda ref: ref is not None, references_iter
        )
        logger.info(f"All references from the stream have been got")
        return references_without_none_iter

    def check_deletion_references_from_stream(self):
        """
        Check the deletion of references from stream

        if reference in Board is missing in Stream
        Then archive the job with this reference
        """
        logger.info("Checking the deletion of references from stream...")
        all_references_from_board_list = list(self.get_all_references_from_board())
        all_references_from_stream_list = list(self.get_all_references_from_stream())
        for reference in all_references_from_board_list:
            if reference not in all_references_from_stream_list:
                # Archive the job with this reference
                logger.debug(f"Archiving the job `{reference}`")
                archive_response = self.hrflow_client.job.indexing.archive(
                    self.board_key, reference=reference, is_archive=1
                )
                if archive_response["code"] >= 400:
                    error_message = archive_response["message"]
                    logger.warning(
                        f"Fail to archive the job `{reference}` : {error_message}"
                    )

        logger.info("The deletion of references from stream has been checked")

    def execute(self) -> Optional[Dict[str, Any]]:
        """
        Execute action
        """
        logger.info("Start execution")

        logger.info("Pulling data...")
        input_data = self.pull()
        logger.info("Data has been pulled")

        logger.info("Applying logics...")
        filtered_data = self.apply_logics(input_data)
        logger.info("Logics have been applied")

        # connect each filtered_data to the format accepted by the pull function (destination, source, board)
        logger.info("Mapping format function...")
        formated_data = map(self.format_switcher, filtered_data)
        logger.info("Format function has been mapped")

        logger.info(
            f"Archive the deleted jobs from stream : {self.archive_deleted_jobs_from_stream}"
        )
        if self.archive_deleted_jobs_from_stream:
            self.check_deletion_references_from_stream()

        logger.info("Filtering the job to push")
        unique_data_to_push = filter(self.check_reference_in_board, formated_data)

        logger.info(f"Hydrate job with parsing : {self.hydrate_with_parsing}")
        if self.hydrate_with_parsing:
            logger.info("Mapping hydrate_job_with_parsing function...")
            unique_data_to_push = map(
                self.hydrate_job_with_parsing, unique_data_to_push
            )
            logger.info("hydrate_job_with_parsing function has been mapped")

        logger.info("Pushing data...")
        self.push(unique_data_to_push)
        logger.info("Data has been pushed")

        logger.info("All has been done for this connector !")


class ProfileDestinationAction(Action):
    hrflow_client: Hrflow = Field(
        ...,
        description="Hrflow client instance used to communicate with the Hrflow.ai API",
    )
    profile: Profile = Field(..., description="Profile to push")

    def pull(self) -> Iterator[TalentDataType]:
        """
        Pull data
        """
        logger.info(
            f"Pulling a profile key=`{self.profile.key}` from a Source `{self.profile.source.key}`"
        )
        response = self.hrflow_client.profile.indexing.get(
            source_key=self.profile.source.key, key=self.profile.key
        )
        if response["code"] >= 400:
            logger.error("Fail to pull a profile")
            raise RuntimeError(
                "Indexing profile get failed : `{}`".format(response["message"])
            )

        profile = response["data"]
        return [profile]

    def execute(self):
        super().execute()
        return generate_workflow_response(
            status_code=201, message="Profile successfully pushed"
        )


from ....core.action import BoardAction
from ....core.http import HTTPStream


from pydantic import Field
import xml.etree.ElementTree
from typing import Iterator


class XMLBoardAction(BoardAction, HTTPStream):
    xml_stream_url: str = Field(..., description="URL to XML Stream")
    job_list_xpath: str = Field(
        ..., description="XPath pointing to the job list in the XML stream"
    )

    @property
    def base_url(self):
        return self.xml_stream_url

    @property
    def http_method(self):
        return "GET"

    def pull(self) -> Iterator[xml.etree.ElementTree.Element]:
        response = self.send_request()

        if response.status_code >= 400:
            error_message = "Unable to pull the data ! Reason : `{}`"
            raise ConnectionError(error_message.format(response.content))

        xml_stream = response.content
        root_element = xml.etree.ElementTree.fromstring(xml_stream)
        job_list_element = root_element.find(self.job_list_xpath)
        job_list = list(job_list_element)
        return job_list


from hrflow import Hrflow
import requests


def workflow(body: dict, settings: dict)-> None:
    """
    CATCH WORKFLOW allows you to run a code function given an API POST request
    @param body: POST request Body
    @rtype: null
    @param settings: dictionary of settings params of the workflow
    """



from builtins import object
from hrflow import Hrflow
import requests

# TODO: verify webhook signature
"""
import hmac
import hashlib
def check_signature(request_signature, secret_key, request_body):
    hasher = hmac.new(secret_key, request_body, hashlib.sha256)
    dig = hasher.hexdigest()
    return hmac.compare_digest(dig, request_signature)
"""


class WebhookEvent(object):
    """
    CATCH Webhook Event
    #TODO signature verification
    """

    def __init__(self, data):
        self.type = data.get("type")
        self.origin = data.get("origin")
        self.message = data.get("message")
        self.profile = data.get("profile")
        if self.profile:
            self.profile_key = self.profile["key"]
            self.source_key = self.profile["source"]["key"]


def workflow(body: dict, settings: dict) -> None:
    """
    CATCH WORKFLOW allows you to run a code function given an API POST request
    @rtype: null
    @param body: POST request Body
    @param settings: dictionary of settings params of the workflow
    """
    event = WebhookEvent(body)
    if not event.profile:
        return
    assert event.type == "profile.parsing.success"  # This workflow is only to create a new profile
    hrflow_client = Hrflow(api_secret=settings["API_KEY"], api_user=settings["USER_EMAIL"])
    crosstalent_api_url = "https://{}.salesforce.com/services/oauth2/token".format(settings["CT_ENV"])
    # Get Crosstalent Token
    try:
        payload = {
            "grant_type": "password",
            "client_id": settings["CT_CLIENT_ID"],
            "client_secret": settings["CT_CLIENT_SECRET"],
            "username": settings["CT_USERNAME"],
            "password": settings["CT_PASSWORD"]
        }
        crosstalent_token = requests.post(crosstalent_api_url, data=payload).json()["access_token"]
    except requests.exceptions.RequestException:
        raise Exception("Retrieving token from Crosstalent failed")
    # Get HrFlow.ai Profile
    try:
        profile_hrflow = hrflow_client.profile.indexing.get(source_key=event.source_key, key=event.profile_key).get(
            "data")
    except requests.exceptions.RequestException:
        raise Exception("Retrieving profile with profile_key: %s and source_key: %s failed" % (event.profile_key, event.source_key))
    # Create HrFlow.ai Profile in Crosstalent
    # TODO: verify that the profile doesn"t exist yet in Crosstalent
    try:
        headers = {
            "Authorization": "OAuth " + crosstalent_token,
            "Content-Type": "application/json"
        }
        crosstalent_profile_endpoint = "https://{}.salesforce.com/services/apexrest/HrFlowCreateProfile/".format(
            settings["CT_URL"])
        r = requests.post(crosstalent_profile_endpoint,
                          json=profile_hrflow,
                          headers=headers)
    except requests.exceptions.RequestException:
        raise Exception("Saving profile with profile_key: %s and source_key: %s failed" % (event.profile_key, event.source_key))


import pytest

from hrflow_connectors.core.auth import XSmartTokenAuth
from hrflow_connectors.connectors.boards.smartrecruiters import SmartJobs


@pytest.fixture
def auth(credentials):
    auth = XSmartTokenAuth(
        value=credentials["smartrecruiters"]["oauth2"]["X-SmartToken"]
    )
    return auth


def test_SmartJobs(logger, auth, hrflow_client):
    action = SmartJobs(
        auth=auth,
        hrflow_client=hrflow_client("dev-demo"),
        limit=2,
        board_key="8ebdea98768dfc04d15f76afab70415ed280ea90",
        hydrate_with_parsing=False,
        archive_deleted_jobs_from_stream=False,
        posting_status="PUBLIC",
    )
    action.execute()



from hrflow import Hrflow
import requests
from datetime import datetime


def hydrate(smart_job):
    return {
        "name": smart_job.get("title"),
        "agent_key": None,
        "reference": smart_job.get("refNumber"),
        "url": None,
        "created_at": smart_job.get("createdOn"),
        "updated_at": smart_job.get("updatedOn"),
        "summary": "",
        "location": {
            "text": " ".join(smart_job.get("location")[key] for key in ["country", "region", "city", "address"]
                             if smart_job.get("location").get(key)),
            "lat": smart_job.get("location").get("latitude"),
            "lng": smart_job.get("location").get("longitude")},
        "sections": [{
            "name": "company_description",
            "title": "companyDescription",
            "description": smart_job.get("jobAd").get("sections").get("companyDescription").get("text")
        },
            {
                "name": "job_description",
                "title": "jobDescription",
                "description": smart_job.get("jobAd").get("sections").get("jobDescription").get("text")
            },
            {
                "name": "qualifications",
                "title": "qualifications",
                "description": smart_job.get("jobAd").get("sections").get("qualifications").get("text")
            },
            {
                "name": "additional_information",
                "title": "additionalInformation",
                "description": smart_job.get("jobAd").get("sections").get("additionalInformation").get("text")
            }
        ],
        "skills": [],
        "languages": [{
            "name": smart_job.get("jobAd").get("language").get("label"),
            "value": None
        }],
        "tags": [{
            "name": "status",
            "value": smart_job.get("status")
        },
            {
                "name": "posting_status",
                "value": smart_job.get("postingStatus")
            },
            {
                "name": "job_uuid",
                "value": smart_job.get("id")
            },
            {
                "name": "experience_level",
                "value": smart_job.get("experienceLevel", {}).get("id")
            },
            {
                "name": "type_of_employment",
                "value": smart_job.get("typeOfEmployment", {}).get("id")
            },
            {
                "name": "industry",
                "value": smart_job.get("industry", {}).get("id")
            },
            {
                "name": "creator",
                "value": xstr(smart_job.get("creator").get("firstName")) + " " + xstr(
                    smart_job.get("creator").get("lastName"))
            },
            {
                "name": "function",
                "value": smart_job.get("function", {}).get("id")},
            {
                "name": "department",
                "value": smart_job.get("department", {}).get("id")},
            {
                "name": "manual",
                "value": smart_job.get("location", {}).get("manual")},
            {
                "name": "remote",
                "value": smart_job.get("location", {}).get("remote")},
            {
                "name": "eeo_category",
                "value": smart_job.get("eeoCategory", {}).get("id")}
        ],
        "ranges_date": [{
            "name": "targetHiringDate",
            "value_min": None,
            "value_max": smart_job.get("targetHiringDate")
        }],
        "ranges_float": [{
            "name": "compensation",
            "value_min": smart_job.get("compensation", {}).get("min"),
            "value_max": smart_job.get("compensation", {}).get("max"),
            "unit": smart_job.get("compensation", {}).get("currency")
        }],
        "metadatas": [],
    }


##### Useful functions
def format_skills(job_description, ents):
    _skills = []
    for ent in ents:
        if ent.get("label") == "HardSkill":
            skill = {
                "name": job_description[ent.get("start"):ent.get("end")],
                "value": None,
                "type": "hard"
            }
            _skills.append(skill)
        if ent.get("label") == "SoftSkill":
            skill = {
                "name": job_description[ent.get("start"):ent.get("end")],
                "value": None,
                "type": "soft"
            }
            _skills.append(skill)
    return _skills


def deduplicate_list(list_to_process):
    deduplicated_list = []
    if len(list_to_process) == 0:
        return []
    if isinstance(list_to_process[0], dict):
        for skill in list_to_process:
            if skill["name"].lower() not in [s["name"] for s in deduplicated_list]:
                deduplicated_list.append({"name": skill["name"].lower(), "value": None, "type": skill["type"]})
        return deduplicated_list
    return list(set(list_to_process))


xstr = lambda s: s or ""


def workflow(settings):
    print("HrFlow.ai client")
    api_secret = settings["API_KEY"]
    api_user = settings["USER_EMAIL"]
    hrflow_client = Hrflow(api_secret=api_secret, api_user=api_user)

    agent_key = settings["AGENT_KEY"]

    token = settings["TOKEN"]

    updated_after = settings["UPDATED_AFTER"]
    offset = _offset = int(settings["OFFSET_JOBS"])

    headers = {
        "X-SmartToken": token
    }

    params = {
        "postingStatus": "PUBLIC",
        "limit": int(settings["LIMIT"]),
        "offset": offset
    }
    if updated_after:
        params["updatedAfter"] = updated_after

    resp = requests.get(url="https://api.smartrecruiters.com/jobs", params=params, headers=headers).json()

    total_found = resp["totalFound"]

    try:
        while _offset < total_found:
            params.update({"offset": _offset})
            resp_jobs = requests.get(url="https://api.smartrecruiters.com/jobs",
                                     params=params,
                                     headers=headers).json()
            jobs = resp_jobs["content"]

            for job in jobs:
                resp_job = requests.get(url="https://api.smartrecruiters.com/jobs/" + job.get("id"),
                                        headers=headers).json()

                _job = hydrate(resp_job)
                job_hrflow = hrflow_client.job.indexing.get(board_key=settings["BOARD_KEY"],
                                                            reference=_job["reference"]).get("data")
                if job_hrflow:
                    job_hrflow["agent_key"] = agent_key
                    job_hrflow.update(_job)

                    print("Index job")
                    resp = hrflow_client.job.indexing.edit(board_key=settings["BOARD_KEY"], key=job_hrflow["key"],
                                                           job_json=job_hrflow)

                    if resp["code"] != 200:
                        raise Exception("Job is not edited")
                else:
                    _job["agent_key"] = agent_key
                    # Compute skills
                    text = "".join(xstr(section["description"]) for section in _job.get("sections"))
                    response = hrflow_client.document.parsing.post(text=text)
                    _job["skills"] = deduplicate_list(format_skills(text, response.get("data").get("ents")))
                    print("Post job")
                    post_response = hrflow_client.job.indexing.add_json(board_key=settings["BOARD_KEY"], job_json=_job)

                    if post_response["code"] != 201:
                        raise Exception("Job is not created")

                _offset += int(settings["LIMIT"])
                updated_after = datetime.utcnow().isoformat() + "Z"

    except Exception as e:
        print(e)
    settings["UPDATED_AFTER"] = updated_after
    return


import logging
import sys
from typing import Union

LOGGER_NAME = "hrflow_connectors"


def get_logger() -> logging.Logger:
    """
    Get logger with `NullHandler` by default

    Returns:
        logging.Logger: logger
    """
    logger = logging.getLogger(LOGGER_NAME)
    if not logger.hasHandlers():
        logger.addHandler(logging.NullHandler())
    return logger


def get_logger_with_basic_config(
    level: Union[int, str] = logging.DEBUG
) -> logging.Logger:
    """
    Get logger with basic configuration :
    * `StreamHandler` on `stdout`
    * Level = `DEBUG`
    * Formatter = `[%(levelname)s][%(asctime)s][%(module)s:%(funcName)s:%(lineno)d] %(message)s`

    Args:
        level (Union[int, str], optional): Log level. Defaults to `logging.DEBUG`.

    Returns:
        logging.Logger: logger with basic configuration
    """
    logger = get_logger()
    logger.setLevel(level)

    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setLevel(level)

    formatter = logging.Formatter(
        "[%(levelname)s][%(asctime)s][%(module)s:%(funcName)s:%(lineno)d] %(message)s"
    )
    stdout_handler.setFormatter(formatter)

    logger.addHandler(stdout_handler)

    return logger